from boto3.dynamodb.conditions import Attr
from boto3.exceptions import ResourceNotExistsError
from typing import List, Dict, Optional

from pydantic import ValidationError

from inoft_vocal_engine.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, PrimaryIndex, Response
from inoft_vocal_engine.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_engine.vocal_apps_model_schemas.models import InoftVocalEngineModelSchema


class VocalAppsModelSchemasDynamoDBClient(DynamoDbCoreAdapter):
    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="accountProjectId", hash_key_variable_python_type=str)
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index, create_table=True)

    def set_update_one_intent(self, accountProjectId: str, intent_model: InoftVocalEngineModelSchema.IntentModel) -> Response:
        response = self._add_update_data_element_to_map(
            key_name="accountProjectId", key_value=accountProjectId,
            object_path_elements={"intents": dict, intent_model.intentId: dict},
            element_values=intent_model.dict()
        )
        return response

    def put_model_schema(self, account_project_id: str, model_schema: InoftVocalEngineModelSchema) -> Response:
        from inoft_vocal_engine.vocal_apps_model_schemas.models import InoftVocalEngineModelSchemaDatabaseComplete
        model_schema = InoftVocalEngineModelSchemaDatabaseComplete(
            accountProjectId=account_project_id, **model_schema.dict()
        )
        response = self._put_item_dict(model_schema.dict())
        return response

    def query_all_intents_by_account_project_id(self, account_project_id: str) -> dict:
        response = self._query_by_key(key_name="accountProjectId", key_value=account_project_id,
                                      fields_to_get=["intents"], query_limit=1)
        if response.count == 1:
            if "intents" in response.items[0]:
                return response.items[0]["intents"]
            else:
                return dict()
        elif response.count > 1:
            print(f"WARNING ! When calling the function {self.query_all_intents_by_account_project_id}, "
                  f"more than one items has been returned with the same accountProjectId."
                  f"  --accountProjectId:{account_project_id}"
                  f"  --response.items:{response.items}"
                  f"  --response.count:{response.count}")
        return dict()

    def get_all_intents_instances_by_account_project_id(self, account_project_id: str) -> List[InoftVocalEngineModelSchema.IntentModel]:
        output_intents_list: List[InoftVocalEngineModelSchema.IntentModel] = list()

        response_intents_dict = self.query_all_intents_by_account_project_id(account_project_id=account_project_id)
        for intent_values in response_intents_dict.values():
            try:
                output_intents_list.append(InoftVocalEngineModelSchema.IntentModel(**intent_values))
            except ValidationError as e:
                print(f"Validation Error while validating an IntentModel from the database : {e}")

        return output_intents_list

    def get_all_intents_dicts_by_account_project_id(self, account_project_id: str) -> List[dict]:
        output_intents_list: List[dict] = list()

        response_intents_dict = self.query_all_intents_by_account_project_id(account_project_id=account_project_id)
        for intent_values in response_intents_dict.values():
            try:
                output_intents_list.append(InoftVocalEngineModelSchema.IntentModel(**intent_values).dict())
            except ValidationError as e:
                print(f"Validation Error while validating an IntentModel from the database : {e}")

        return output_intents_list

    def update_value_with_path_target(self, account_project_id: str, object_path_elements: Dict[str, type], value: any):
        update_item_response = self._add_update_data_element_to_map(
            key_name="accountProjectId", key_value=account_project_id,
            object_path_elements=object_path_elements, element_values=value
        )
        return update_item_response

    def get_item_in_path_target(self, account_project_id: str, target_path_elements: Dict[str, type]) -> Optional[dict]:
        return self._get_item_in_path_target(
            key_name="accountProjectId", key_value=account_project_id,
            target_path_elements=target_path_elements
        )



from inoft_vocal_engine.speech_synthesis.polly.voice_object import Voice

French_France_Female_CELINE = Voice(gender='Female', id='Celine', language_code='fr-FR', language_name='French', name='Céline', supported_engines=['standard'])
French_France_Male_MATHIEU = Voice(gender='Male', id='Mathieu', language_code='fr-FR', language_name='French', name='Mathieu', supported_engines=['standard'])
French_France_Female_LEA = Voice(gender='Female', id='Lea', language_code='fr-FR', language_name='French', name='Léa', supported_engines=['standard'])

Dutch_Netherlands_Female_LOTTE = Voice(gender='Female', id='Lotte', language_code='nl-NL', language_name='Dutch', name='Lotte', supported_engines=['standard'])
Russian_Russia_Male_MAXIM = Voice(gender='Male', id='Maxim', language_code='ru-RU', language_name='Russian', name='Maxim', supported_engines=['standard'])
English_UnitedStates_Female_SALLI = Voice(gender='Female', id='Salli', language_code='en-US', language_name='US English', name='Salli', supported_engines=['standard'])
English_Welsh_Male_GERAINT = Voice(gender='Male', id='Geraint', language_code='en-GB-WLS', language_name='Welsh English', name='Geraint', supported_engines=['standard'])
Spanish_UnitedStates_Male_MIGUEL = Voice(gender='Male', id='Miguel', language_code='es-US', language_name='US Spanish', name='Miguel', supported_engines=['standard'])
German_Germany_Female_MARLENE =Voice(gender='Female', id='Marlene', language_code='de-DE', language_name='German', name='Marlene', supported_engines=['standard'])
Italian_Italy_Male_GIORGIO = Voice(gender='Male', id='Giorgio', language_code='it-IT', language_name='Italian', name='Giorgio', supported_engines=['standard'])
Portuguese_Portugal_Female_INES = Voice(gender='Female', id='Ines', language_code='pt-PT', language_name='Portuguese', name='Inês', supported_engines=['standard'])
Arabic_Female_ZEINA = Voice(gender='Female', id='Zeina', language_code='arb', language_name='Arabic', name='Zeina', supported_engines=['standard'])
ChineseMandarin_China_ZHIYU = Voice(gender='Female', id='Zhiyu', language_code='cmn-CN', language_name='Chinese Mandarin', name='Zhiyu', supported_engines=['standard'])
Welsh_Female_GWYNETH = Voice(gender='Female', id='Gwyneth', language_code='cy-GB', language_name='Welsh', name='Gwyneth', supported_engines=['standard'])
Icelandic_Iceland_Male_KARL = Voice(gender='Male', id='Karl', language_code='is-IS', language_name='Icelandic', name='Karl', supported_engines=['standard'])
English_UnitedStates_Female_JOANNA = Voice(gender='Female', id='Joanna', language_code='en-US', language_name='US English', name='Joanna', supported_engines=['standard'])
CastalianSpanish_Spain_Female_LUCIA = Voice(gender='Female', id='Lucia', language_code='es-ES', language_name='Castilian Spanish', name='Lucia', supported_engines=['standard'])
Portuguese_Portugal_Male_CRISTIANO = Voice(gender='Male', id='Cristiano', language_code='pt-PT', language_name='Portuguese', name='Cristiano', supported_engines=['standard'])
Swedish_Sweden_Female_ASTRID = Voice(gender='Female', id='Astrid', language_code='sv-SE', language_name='Swedish', name='Astrid', supported_engines=['standard'])
German_Germany_Female_VICKY = Voice(gender='Female', id='Vicki', language_code='de-DE', language_name='German', name='Vicki', supported_engines=['standard'])

# todo: finish the references of the below voices
""""
Voice(gender='Female', id='Mia', language_code='es-MX', language_name='Mexican Spanish', name='Mia', supported_engines=['standard'])
Voice(gender='Female', id='Bianca', language_code='it-IT', language_name='Italian', name='Bianca', supported_engines=['standard'])
Voice(gender='Female', id='Vitoria', language_code='pt-BR', language_name='Brazilian Portuguese', name='Vitória', supported_engines=['standard'])
Voice(gender='Female', id='Raveena', language_code='en-IN', language_name='Indian English', name='Raveena', supported_engines=['standard'])
Voice(gender='Female', id='Chantal', language_code='fr-CA', language_name='Canadian French', name='Chantal', supported_engines=['standard'])
Voice(gender='Female', id='Amy', language_code='en-GB', language_name='British English', name='Amy', supported_engines=['standard'])
Voice(gender='Male', id='Brian', language_code='en-GB', language_name='British English', name='Brian', supported_engines=['standard'])
Voice(gender='Male', id='Russell', language_code='en-AU', language_name='Australian English', name='Russell', supported_engines=['standard'])
Voice(gender='Female', id='Aditi', language_code='en-IN', language_name='Indian English', name='Aditi', 'AdditionalLanguageModels': ['hi-IN'], supported_engines=['standard'])
Voice(gender='Male', id='Matthew', language_code='en-US', language_name='US English', name='Matthew', supported_engines=['standard'])
Voice(gender='Female', id='Dora', language_code='is-IS', language_name='Icelandic', name='Dóra', supported_engines=['standard'])
Voice(gender='Male', id='Enrique', language_code='es-ES', language_name='Castilian Spanish', name='Enrique', supported_engines=['standard'])
Voice(gender='Male', id='Hans', language_code='de-DE', language_name='German', name='Hans', supported_engines=['standard'])
Voice(gender='Female', id='Carmen', language_code='ro-RO', language_name='Romanian', name='Carmen', supported_engines=['standard'])
Voice(gender='Female', id='Ivy', language_code='en-US', language_name='US English', name='Ivy', supported_engines=['standard'])
Voice(gender='Female', id='Ewa', language_code='pl-PL', language_name='Polish', name='Ewa', supported_engines=['standard'])
Voice(gender='Female', id='Maja', language_code='pl-PL', language_name='Polish', name='Maja', supported_engines=['standard'])
Voice(gender='Female', id='Nicole', language_code='en-AU', language_name='Australian English', name='Nicole', supported_engines=['standard'])
Voice(gender='Female', id='Camila', language_code='pt-BR', language_name='Brazilian Portuguese', name='Camila', supported_engines=['standard'])
Voice(gender='Female', id='Filiz', language_code='tr-TR', language_name='Turkish', name='Filiz', supported_engines=['standard'])
Voice(gender='Male', id='Jacek', language_code='pl-PL', language_name='Polish', name='Jacek', supported_engines=['standard'])
Voice(gender='Male', id='Justin', language_code='en-US', language_name='US English', name='Justin', supported_engines=['standard'])
Voice(gender='Female', id='Celine', language_code='fr-FR', language_name='French', name='Céline', supported_engines=['standard'])
Voice(gender='Female', id='Kendra', language_code='en-US', language_name='US English', name='Kendra', supported_engines=['standard'])
Voice(gender='Male', id='Ricardo', language_code='pt-BR', language_name='Brazilian Portuguese', name='Ricardo', supported_engines=['standard'])
Voice(gender='Male', id='Mads', language_code='da-DK', language_name='Danish', name='Mads', supported_engines=['standard'])
Voice(gender='Male', id='Mathieu', language_code='fr-FR', language_name='French', name='Mathieu', supported_engines=['standard'])
Voice(gender='Female', id='Lea', language_code='fr-FR', language_name='French', name='Léa', supported_engines=['standard'])
Voice(gender='Female', id='Naja', language_code='da-DK', language_name='Danish', name='Naja', supported_engines=['standard'])
Voice(gender='Female', id='Penelope', language_code='es-US', language_name='US Spanish', name='Penélope', supported_engines=['standard'])
Voice(gender='Female', id='Tatyana', language_code='ru-RU', language_name='Russian', name='Tatyana', supported_engines=['standard'])
Voice(gender='Male', id='Ruben', language_code='nl-NL', language_name='Dutch', name='Ruben', supported_engines=['standard'])
Voice(gender='Female', id='Mizuki', language_code='ja-JP', language_name='Japanese', name='Mizuki', supported_engines=['standard'])
Voice(gender='Male', id='Takumi', language_code='ja-JP', language_name='Japanese', name='Takumi', supported_engines=['standard'])
Voice(gender='Female', id='Conchita', language_code='es-ES', language_name='Castilian Spanish', name='Conchita', supported_engines=['standard'])
Voice(gender='Female', id='Carla', language_code='it-IT', language_name='Italian', name='Carla', supported_engines=['standard'])
Voice(gender='Female', id='Kimberly', language_code='en-US', language_name='US English', name='Kimberly', supported_engines=['standard'])
Voice(gender='Male', id='Jan', language_code='pl-PL', language_name='Polish', name='Jan', supported_engines=['standard'])
Voice(gender='Female', id='Liv', language_code='nb-NO', language_name='Norwegian', name='Liv', supported_engines=['standard'])
Voice(gender='Male', id='Joey', language_code='en-US', language_name='US English', name='Joey', supported_engines=['standard'])
Voice(gender='Female', id='Lupe', language_code='es-US', language_name='US Spanish', name='Lupe', supported_engines=['standard'])
Voice(gender='Female', id='Seoyeon', language_code='ko-KR', language_name='Korean', name='Seoyeon', supported_engines=['standard'])
Voice(gender='Female', id='Emma', language_code='en-GB', language_name='British English', name='Emma', supported_engines=['standard'])])
"""




from dataclasses import dataclass
from typing import Optional


@dataclass
class PermissionModel:
    expression: str
    display_name: str
    description: Optional[str] = None
    additional_kwargs: Optional[dict] = None



class Voice:
    def __init__(self, gender: str, id: str, language_code: str, language_name: str, name: str, supported_engines: list or str):
        self.id = id
        self.gender = gender
        self.name = name
        self.language_code = language_code
        self.language_name = language_name
        self.supported_engines = supported_engines



import time
from dataclasses import asdict
from typing import Any, Optional, List, Dict, Tuple
from boto3.exceptions import ResourceNotExistsError
from boto3.dynamodb.conditions import Attr, Key
from pydantic import BaseModel, validator, ValidationError

from inoft_vocal_engine.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, PrimaryIndex, GlobalSecondaryIndex, Response
from inoft_vocal_engine.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.auth.models import AuthToken, AccountData


class InstanceInfoDatabaseItem(BaseModel):
    projectId: str
    instanceName: str
    instanceDescription: Optional[str] = None
    instanceId: Optional[str] = None

    @validator('instanceId', always=True)
    def make_uuid_if_missing(cls, value):
        if value is not None:
            return value
        else:
            from uuid import uuid4
            instance_id = str(uuid4())
            print(f"Initializing new instance --instanceId:{instance_id}")
            return instance_id

    def to_database(self) -> dict:
        return self.dict(exclude={"projectId"})

class ProjectDatabaseItem(BaseModel):
    """ To be used in the projects dict of the AccountDatabaseItem """
    projectName: str
    projectPrimaryUrl: str
    projectDescription: Optional[str] = None
    instancesInfos: Optional[Dict[str, InstanceInfoDatabaseItem]] = dict()
    accountProjectId: Optional[str] = None

    @validator('accountProjectId', always=True)
    def make_uuid_if_missing(cls, value):
        if value is not None:
            return value
        else:
            from uuid import uuid4
            account_project_id = str(uuid4())
            print(f"Initializing new project --accountProjectId:{account_project_id}")
            return account_project_id

class AccountDatabaseItem(BaseModel):
    email: str
    username: str
    encryptedPassword: str
    accountId: Optional[str] = None
    authTokens: Optional[list] = list()
    projects: Optional[dict] = dict()
    permissionsGrantedToOtherAccounts: Optional[Dict[str, List[str]]] = dict()
    permissionsGivenToAccount: Optional[Dict[str, List[str]]] = dict()
    projectsUrlsToIds: Optional[Dict[str, str]] = list()
    sharedProjectsIds: Optional[List[str]] = list()
    # todo: remove use of fields that were used when this class was a dataclass instead of a BaseModel

    @validator('accountId', always=True)
    def make_uuid_if_missing(cls, value):
        if value is not None:
            return value
        else:
            from uuid import uuid4
            account_id = str(uuid4())
            print(f"Initializing new account --accountId:{account_id}")
            return account_id

class AWSBuildInfosDatabaseModel(BaseModel):
    provider: str = "AWS"
    buildName: str = "development"
    lambdaArn: Optional[str] = None
    apiGatewayId: Optional[str] = None

class AWSBuildInfosBackendModel(AWSBuildInfosDatabaseModel):
    buildId: str

    def to_database_model(self) -> AWSBuildInfosDatabaseModel:
        return AWSBuildInfosDatabaseModel(
            provider=self.provider,
            buildName=self.buildName,
            lambdaArn=self.lambdaArn,
            apiGatewayId=self.apiGatewayId
        )



class UsersDynamoDBClient(DynamoDbCoreAdapter):
    AUTH_TOKEN_SECONDS_TIMEOUT = 86400  # 24 hours in seconds

    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="accountId", hash_key_variable_python_type=str)
        globals_secondary_indexes = [
            GlobalSecondaryIndex(hash_key_name="username", hash_key_variable_python_type=str, projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="email", hash_key_variable_python_type=str, projection_type="ALL"),
        ]
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index,
                         global_secondary_indexes=globals_secondary_indexes, create_table=True)
        # The table should exist in only one copy, which is why its name and region is hardcoded

    def put_new_account(self, account_item: AccountDatabaseItem) -> AccountDatabaseItem:
        account_item_dict = asdict(account_item)
        self._put_item_dict(item_dict=account_item_dict)
        print(f"Finished creation of new account."
              f"  --item{account_item_dict}"
              f"  --time:{time.time()}")
        return account_item

    def add_project_to_account(self, account_id: str, project_item: ProjectDatabaseItem,
                               project_secondary_urls: Optional[List[str]] = None):
        # todo: before creating a project, check if no project in the account has the same primary url

        project_item_dict_values = project_item.dict()
        account_item_set_response = self._add_update_data_element_to_map(
            key_name="accountId", key_value=account_id,
            object_path_elements={'projects': dict, project_item.accountProjectId: str},
            element_values=project_item_dict_values
        )
        if account_item_set_response is None:
            print(f"Error while adding new project item to an account."
                  f"  --account_id:{account_id}"
                  f"  --key:{project_item.projectName}"
                  f"  --item:{project_item_dict_values}"
                  f"  --time:{time.time()}")
            return False
        else:
            if project_secondary_urls is None:
                # If we have only one element to add to a map, the add_data_element_to_map is more efficient in processing power for the server
                # than add_multiple_data_elements_to_map. Otherwise the add_multiple_data_elements_to_map is more efficient in database accesses.
                self._add_update_data_element_to_map(
                    key_name="accountId", key_value=account_id,
                    object_path_elements={'projectsUrlsToIds': dict, project_item.projectName: str},
                    element_values=project_item.accountProjectId
                )
            else:
                from inoft_vocal_engine.databases.dynamodb.dynamodb_core import DynamoDBMapObjectSetter
                project_secondary_urls.append(project_item.projectName)
                project_urls = project_secondary_urls
                objects_setters: List[DynamoDBMapObjectSetter] = list()
                for url in project_urls:
                    objects_setters.append(DynamoDBMapObjectSetter(object_path='projectsUrlsToIds', map_key=url,
                                                                   value_to_set=project_item.accountProjectId))
                self._add_update_multiple_data_elements_to_map(key_name="accountId", key_value=account_id, objects_setters=objects_setters)

            print(f"Added new project item to an account."
                  f"  --account_id:{account_id}"
                  f"  --key:{project_item.projectName}"
                  f"  --item:{project_item_dict_values}"
                  f"  --time:{time.time()}")
            return True

    def add_instance_infos_to_account(self, account_id: str, instance_infos_item: InstanceInfoDatabaseItem) -> bool:
        instance_infos_item_dict_values = instance_infos_item.to_database()
        instance_infos_update_response = self._add_update_data_element_to_map(
            key_name="accountId", key_value=account_id, element_values=instance_infos_item_dict_values,
            object_path_elements={
                'projects': dict, instance_infos_item.projectId: dict,
                'instancesInfos': dict, instance_infos_item.instanceId: dict
            }
        )
        if instance_infos_update_response is not None:
            return True
        else:
            print(f"Error while adding new instance infos item to an account."
                  f"  --account_id:{account_id}"
                  f"  --key:{instance_infos_item.projectId}"
                  f"  --item:{instance_infos_item_dict_values}"
                  f"  --time:{time.time()}")
            return False

    # todo: decide if the password validation should be directly done in the database queries ?
    def get_account_by_username(self, username: str, fields_to_get: list) -> dict:
        return self._query_single_item_by_key(index_name="username", key_name="username", key_value=username, fields_to_get=fields_to_get)

    def get_account_by_email(self, email: str, fields_to_get: list) -> dict:
        return self._query_single_item_by_key(index_name="email", key_name="email", key_value=email, fields_to_get=fields_to_get)

    def check_permissions_on_account_project(self, project_owner_account_username: str, requester_account_id: str,
                                             permissions_to_check: List[str], fields_to_get: List[str]):

        requester_expected_permissions_map_path = f"permissionsGrantedToOtherAccounts.{requester_account_id}"
        # We store this string in a variable, just to not have to create it multiple times in the permissions loop

        filter_expression = Attr(requester_expected_permissions_map_path).exists()
        for permission in permissions_to_check:
            filter_expression = self._add_to_filter_expression(
                expression=filter_expression,
                condition=Attr(requester_expected_permissions_map_path).contains(permission)
            )

        response = self._query_by_key(
            index_name="username", key_name="username", key_value=project_owner_account_username,
            filter_expression=filter_expression, fields_to_get=fields_to_get
        )
        return response

    def get_account_permissions_granted_to_other_accounts_by_username(self, owner_account_username: str,
                                                                      account_id_to_get_permissions_for: str) -> Tuple[dict, bool]:

        response = self._query_for_account(index_name="username", key_name="username", key_value=owner_account_username,
                                           fields_to_get=[f"permissionsGrantedToOtherAccounts.{account_id_to_get_permissions_for}"])
        return response

    def remove_old_auth_tokens(self, account_id: str, fields_to_get: List[str]) -> AccountData:
        # account_data, account_been_found \
        # todo: optimize this code, because currently when logging, we have multiple accesses on the users database
        get_account_response = self._query_by_key(key_name="accountId", key_value=account_id, fields_to_get=fields_to_get)

        if len(get_account_response.items) > 0:
            account_data = AccountData(**get_account_response.items[0])
            indexes_tokens_to_remove = list()

            current_time = time.time()
            if account_data.authTokens is not None:
                for i, auth_token in enumerate(account_data.authTokens):
                    try:
                        if current_time > auth_token.creationTimestamp + self.AUTH_TOKEN_SECONDS_TIMEOUT:
                            indexes_tokens_to_remove.append(i)
                    except Exception as e:
                        indexes_tokens_to_remove.append(i)
                        print(f"Auth toke {auth_token} was invalid and will be removed. It has caused the following error {e}")

            if len(indexes_tokens_to_remove) > 0:
                update_expression = "REMOVE "
                for i, token_index_in_database_list in enumerate(indexes_tokens_to_remove):
                    update_expression += f"authTokens[{token_index_in_database_list}]"
                    if i + 1 < len(indexes_tokens_to_remove):
                        update_expression += ", "

                self._remove_data_elements_from_list(key_name="accountId", key_value=account_id,
                                                     list_object_path="authTokens", indexes_to_remove=indexes_tokens_to_remove)

                for i_old_token in indexes_tokens_to_remove:
                    account_data.authTokens.pop(i_old_token)
            return account_data

    def add_auth_token(self, account_id: str) -> str:
        from uuid import uuid4
        auth_token = AuthToken(token=str(uuid4()), creationTimestamp=int(time.time()))
        self._add_data_elements_to_list(key_name="accountId", key_value=account_id,
                                        object_path="authTokens", element_values=[auth_token.dict()])
        return auth_token.token

    def get_account_project_id_by_project_url(self, project_url: str, account_id: Optional[str] = None, account_username: Optional[str] = None) -> Optional[str]:
        kwargs = {}
        if account_id is not None:
            kwargs["key_name"] = "accountId"
            kwargs["key_value"] = account_id
        elif account_username is not None:
            kwargs["key_name"] = "username"
            kwargs["key_value"] = account_username
            kwargs["index_name"] = "username"

        response_items = self._query_by_key(**kwargs, query_limit=1,
            fields_to_get=[f"projectsUrlsToIds.{project_url}"],
            filter_expression=Attr(f"projectsUrlsToIds.{project_url}").exists()
        ).items

        if len(response_items) > 0:
            account_project_id = SafeDict(response_items[0]).get("projectsUrlsToIds").get(project_url).to_str(default=None)
            return account_project_id

    def get_project_by_project_url(self, account_id: str, project_url: str) -> Optional[ProjectDatabaseItem]:
        account_project_id = self.get_account_project_id_by_project_url(account_id=account_id, project_url=project_url)

        if account_project_id is not None:
            response_items = self._query_by_key(
                key_name="accountId", key_value=account_id,
                fields_to_get=[f"projects.{account_project_id}"],
                filter_expression=Attr(f"projects.{account_project_id}").exists()
            ).items
            if len(response_items) > 0:
                project_dict = SafeDict(response_items[0]).get("projects").get(account_project_id).to_dict(default=None)
                if project_dict is not None:
                    project_instance = ProjectDatabaseItem(**project_dict)
                    return project_instance

    def _get_all_owned_projects_of_account(self, key_name: str, key_value: str, index_name: Optional[str] = None) -> Optional[List[ProjectDatabaseItem]]:
        kwargs = dict()
        if index_name is not None:
            kwargs["index_name"] = index_name

        response_items = self._query_by_key(key_name=key_name, key_value=key_value,
                                            fields_to_get=["projects"], **kwargs).items
        if len(response_items) > 0:
            projects_data_dict = SafeDict(response_items[0]).get("projects").to_dict(default=None)
            if projects_data_dict is not None:
                projects_database_items_instances: List[ProjectDatabaseItem] = list()
                for project_key_id, project_data in projects_data_dict.items():
                    try:
                        projects_database_items_instances.append(ProjectDatabaseItem(**project_data))
                    except ValidationError as e:
                        print(e)
                return projects_database_items_instances

    def get_all_owned_projects_of_account_by_account_id(self, account_id: str) -> Optional[List[ProjectDatabaseItem]]:
        return self._get_all_owned_projects_of_account(key_name="accountId", key_value=account_id)

    def get_all_owned_projects_of_account_by_account_username(self, account_username: str) -> Optional[List[ProjectDatabaseItem]]:
        return self._get_all_owned_projects_of_account(key_name="username", key_value=account_username, index_name="username")

    def get_infos_of_one_project_build(self, project_resources, build_id: str) -> Optional[AWSBuildInfosBackendModel]:
        from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
        project_resources: ProjectResources

        build_infos: Optional[dict] = self._get_value_in_path_target(
            key_name="accountId", key_value=project_resources.project_owner_account_id,
            target_path_elements={"projects": dict, project_resources.account_project_id: dict, "builds": dict, build_id: dict}
        )
        if build_infos is not None:
            try:
                build_infos["buildId"] = build_id
                # We set the buildId in the dict of the build_infos, instead of directly passing the buildId as a
                # kwarg, to avoid issues if for some reasons the buildId kwarg was present in the build_infos dict.
                return AWSBuildInfosBackendModel(**build_infos)
            except ValidationError as e:
                print(e)
        return None

    def set_update_infos_of_one_project_build(self, project_resources, build: AWSBuildInfosBackendModel) -> bool:
        from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
        project_resources: ProjectResources

        response = self._add_update_data_element_to_map(
            key_name="accountId", key_value=project_resources.project_owner_account_id,
            object_path_elements={"projects": dict, project_resources.account_project_id: dict, "builds": dict, build.buildId: dict},
            element_values=build.to_database_model().dict()
        )
        if response is not None:
            return True
        return False





import os
from dataclasses import field

from ask_smapi_model.v1.skill.manifest import SkillManifest
from pydantic import BaseModel, validator, constr
from typing import List, Dict, Optional


class InoftVocalEnginePublishingInfos(BaseModel):
    class AlexaSpecificModel(BaseModel):
        class LocalizedTabDataModel(BaseModel):
            alexaSkillName: Optional[str]
            alexaSkillDescription: Optional[str] = ""
            alexaSkillSummary: Optional[str] = ""
            alexaTermsOfUseUrl: Optional[str] = ""
            alexaPrivacyPolicyUrl: Optional[str] = ""
            alexaSkillExamplePhrases: Optional[List[str]] = list()
            alexaSkillKeywords: Optional[List[str]] = list()

            @validator("alexaSkillName")
            def check_length_of_name(cls, value):
                if value is not None and isinstance(value, str) and len(value) > 1:
                    return value
                return "Temporary name"

        selectedCountriesCode: List[str]
        selectedSkillCategory: Optional[str] = ""
        testInstructionsForAmazon: Optional[str] = ""
        localizedTabsExportedData: Dict[str, LocalizedTabDataModel]

        @validator("selectedSkillCategory")
        def check_if_skill_category_valid(cls, value):
            if value is None or value == "":
                return value

            from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path
            skill_categories_json_filepath = os.path.join(get_inoft_vocal_engine_root_path(), "web_interface/static/deployment/alexaSkillCategories.json")

            from inoft_vocal_engine.utils.general import load_json
            valid_categories = load_json(filepath=skill_categories_json_filepath)

            if value in valid_categories:
                return value
            else:
                raise ValueError(f"The category with the key {value} was not found in the valid categories json file : {valid_categories}")

    alexaSpecific: AlexaSpecificModel

    def to_alexa_manifest(self) -> SkillManifest:
        from ask_smapi_model.v1.isp import PublishingInformation
        from ask_smapi_model.v1.isp import PrivacyAndCompliance
        from ask_smapi_model.v1.skill.manifest import SkillManifestEvents
        from ask_smapi_model.v1.skill.manifest import PermissionItems
        from ask_smapi_model.v1.skill.manifest import SkillManifestApis
        from ask_smapi_model.v1.skill.alexa_hosted import HostingConfiguration
        from ask_smapi_model.v1.skill.manifest import DistributionCountries
        from ask_smapi_model.v1.skill.manifest import CustomApis
        from ask_smapi_model.v1.skill.manifest import SkillManifestEndpoint

        from ask_smapi_model.v1.skill.manifest import SkillManifestPublishingInformation
        from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPublishingInformation
        from ask_smapi_model.v1.skill.manifest import SkillManifestPrivacyAndCompliance
        alexa_manifest = SkillManifest(
            manifest_version="1.0",
            publishing_information=SkillManifestPublishingInformation(
                locales={},
                distribution_countries=self.alexaSpecific.selectedCountriesCode,
                is_available_worldwide=False,
                testing_instructions=self.alexaSpecific.testInstructionsForAmazon,
                category=self.alexaSpecific.selectedSkillCategory
            ),
            privacy_and_compliance=SkillManifestPrivacyAndCompliance(
                locales={},
                allows_purchases=False,
                is_export_compliant=True,
                is_child_directed=False,
                uses_personal_info=False
            ),
            apis={
                "custom": {
                    "endpoint": {
                        "uri": "arn:aws:lambda:us-east-1:032174894474:function:ask-custom-custome_cert"
                    }
                }
            }
            # events=SkillManifestEvents(),
            # permissions=[PermissionItems()],
            # apis=SkillManifestApis(
            #    custom=CustomApis(
            #        endpoint=SkillManifestEndpoint(
            #            uri="arn:aws:lambda:eu-west-3:631258222318:function:testlambdacoolose"
            #        )
            #    )
            # )
        )

        for localLanguageKeyName, localInfosData in self.alexaSpecific.localizedTabsExportedData.items():
            from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPublishingInformation
            alexa_manifest.publishing_information.locales[localLanguageKeyName] = (
                SkillManifestLocalizedPublishingInformation(
                    name=localInfosData.alexaSkillName,
                    description=localInfosData.alexaSkillDescription,
                    summary=localInfosData.alexaSkillSummary,
                    example_phrases=localInfosData.alexaSkillExamplePhrases,
                    keywords=localInfosData.alexaSkillKeywords
                )
            )

            from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPrivacyAndCompliance
            alexa_manifest.privacy_and_compliance.locales[localLanguageKeyName] = (
                SkillManifestLocalizedPrivacyAndCompliance(
                    privacy_policy_url=localInfosData.alexaPrivacyPolicyUrl,
                    terms_of_use_url=localInfosData.alexaTermsOfUseUrl
                )
            )
            # todo: check if the urls are valid http or https url, otherwise
            #  there will be an internal client error when sending the data
        # serialized_manifest = nested_serialize(alexa_manifest)
        return alexa_manifest



from inoft_vocal_framework.exceptions import raise_if_variable_not_expected_type


class Endpoints:
    def __init__(self, alexa_api_endpoint_url_not_recommended_to_use: str = None,
                 google_assistant_api_endpoint_url: str = None, samsung_bixy_api_endpoint_url: str = None,
                 siri_api_endpoint_url: str = None):
        self.alexa_api_endpoint_url_not_recommended_to_use = None
        self.googleAssistantApiEndointUrl = None
        self.samsungBixbyApiEndointUrl = None
        self.siriApiEndointUrl = None


class Deployment:
    def __init__(self, handler_function_path: str, api_gateway_id: str, s3_bucket_name: str, lambda_name: str,
                 endpoints: Endpoints):
        self.handler_function_path = handler_function_path
        self.api_gateway_id = api_gateway_id
        self.s3_bucket_name = s3_bucket_name
        self.lambda_name = lambda_name
        self.endpoints = endpoints

    @property
    def handler_function_path(self) -> str:
        return self._handler_function_path

    @handler_function_path.setter
    def handler_function_path(self, handler_function_path: str) -> None:
        raise_if_variable_not_expected_type(value=handler_function_path, expected_type=str,
                                            variable_name="handler_function_path")
        self._handler_function_path = handler_function_path

    @property
    def api_gateway_id(self) -> str:
        return self._api_gateway_id

    @api_gateway_id.setter
    def api_gateway_id(self, api_gateway_id: str) -> None:
        raise_if_variable_not_expected_type(value=api_gateway_id, expected_type=str, variable_name="api_gateway_id")
        self._api_gateway_id = api_gateway_id

    @property
    def s3_bucket_name(self) -> str:
        return self._s3_bucket_name

    @s3_bucket_name.setter
    def s3_bucket_name(self, s3_bucket_name: str) -> None:
        raise_if_variable_not_expected_type(value=s3_bucket_name, expected_type=str, variable_name="s3_bucket_name")
        self._s3_bucket_name = s3_bucket_name

    @property
    def lambda_name(self) -> str:
        return self._lambda_name

    @lambda_name.setter
    def lambda_name(self, lambda_name: str) -> None:
        raise_if_variable_not_expected_type(value=lambda_name, expected_type=str, variable_name="lambda_name")
        self._lambda_name = lambda_name

    @property
    def endpoints(self) -> Endpoints:
        return self._endpoints

    @endpoints.setter
    def endpoints(self, endpoints: Endpoints) -> None:
        raise_if_variable_not_expected_type(value=endpoints, expected_type=self.Endpoints, variable_name="endpoints")
        self._endpoints = endpoints


"""Getting Started Example for Python 2.7+/3.3+"""
from typing import Optional

from boto3 import Session
from botocore.exceptions import BotoCoreError, ClientError
from contextlib import closing
import os
import sys


class PollyClient:
    def __init__(self):
        session = Session()
        self.client = session.client("polly")

    def synthesize(self, text: str, voice_id: str, filepath_to_save_to: str,
                   output_format: Optional[str] = "mp3", should_play_audio:  Optional[bool] = False):
        try:
            response = self.client.synthesize_speech(Text=text, VoiceId=voice_id, OutputFormat=output_format)
        except (BotoCoreError, ClientError) as error:
            # The service returned an error, exit gracefully
            print(f"Error while synthesising using the voice {voice_id} the following text : {text} - {error}")
            return None

        # Access the audio stream from the response
        if "AudioStream" in response:
            # Note: Closing the stream is important because the service throttles on the
            # number of parallel connections. Here we are using contextlib.closing to
            # ensure the close method of the stream object will be called automatically
            # at the end of the with statement's scope.
            with closing(response["AudioStream"]) as stream:
                try:
                    # Open a file for writing the output as a binary stream
                    with open(filepath_to_save_to, "wb+") as file:
                        file.write(stream.read())
                except IOError as error:
                    # Could not write to file, exit gracefully
                    print(f"Error while saving the synthesized speech with the voice {voice_id} of the following text : {text} - {error}")
                    return None

        else:
            # The response didn't contain audio data, exit gracefully
            print(f"AWS Polly did not send AudioDate while synthesizing using the voice {voice_id} the following text : {text}")
            return None

        if should_play_audio is True:
            # Play the audio using the platform's default player
            if sys.platform == "win32":
                os.startfile(filepath_to_save_to)
            else:
                # The following works on macOS and Linux. (Darwin = mac, xdg-open = linux).
                import subprocess
                opener = "open" if sys.platform == "darwin" else "xdg-open"
                subprocess.call([opener, filepath_to_save_to])

        return filepath_to_save_to


if __name__ == "__main__":
    from inoft_vocal_engine.speech_synthesis.polly import VOICES

    PollyClient().synthesize(text="Bien le bonjour l'ami. T'aime le chocolat ?", voice_id=VOICES.Icelandic_Iceland_Male_KARL.id,
                             filepath_to_save_to="F:/Inoft/skill_histoire_decryptage_1/inoft_vocal_engine/speech_synthesis/polly/test.mp3",
                             should_play_audio=True)



import pyaudio
import wave

chunk = 1024  # Record in chunks of 1024 samples
sample_format = pyaudio.paInt16  # 16 bits per sample
channels = 2
fs = 44100  # Record at 44100 samples per second
seconds = 3
filename = "output.wav"

p = pyaudio.PyAudio()  # Create an interface to PortAudio

print('Recording')

stream = p.open(format=sample_format,
                channels=channels,
                rate=fs,
                frames_per_buffer=chunk,
                input=True)

frames = []  # Initialize array to store frames

# Store data in chunks for 3 seconds
for i in range(0, int(fs / chunk * seconds)):
    data = stream.read(chunk)
    frames.append(data)

# Stop and close the stream
stream.stop_stream()
stream.close()
# Terminate the PortAudio interface
p.terminate()

print('Finished recording')

# Save the recorded data as a WAV file
wf = wave.open(filename, 'wb')
wf.setnchannels(channels)
wf.setsampwidth(p.get_sample_size(sample_format))
wf.setframerate(fs)
wf.writeframes(b''.join(frames))
wf.close()


import os


def get_inoft_vocal_engine_root_path() -> os.path:
    import inoft_vocal_engine
    return os.path.dirname(os.path.abspath(inoft_vocal_engine.__file__))

def get_engine_temp_path() -> os.path:
    return os.path.join(get_inoft_vocal_engine_root_path(), "tmp")



from typing import List

from pydub import AudioSegment

from inoft_vocal_engine.audio_editing.relation import Relation
from inoft_vocal_engine.audio_editing.sound import Sound


def apply_relation_settings(sound: Sound) -> Sound:
    if sound.relation.parent_sound.seconds_start_time is None:
        raise Exception(f"The parent sound with source found at {sound.relation.parent_sound.local_filepath} is not being\n"
                        f"rendered before the child sound of the relation with source found at {sound.local_filepath}.\n"
                        f"Please try to set your tracks containing your parent sound before in your AudioClip,\n"
                        f"or in your track try to set your parent sound before adding your child sound\n"
                        f"(the order with which you create relations does not matter)")

    if sound.relation.seconds_child_start_after_parent_start is not None:
        sound.seconds_start_time = sound.relation.parent_sound.seconds_start_time + sound.relation.seconds_child_start_after_parent_start
    elif sound.relation.seconds_child_start_after_parent_end is not None:
        sound.seconds_start_time = sound.relation.parent_sound.seconds_end_time + sound.relation.seconds_child_start_after_parent_end
    elif sound.relation.seconds_child_end_after_parent_start is not None:
        sound.seconds_end_time = sound.relation.parent_sound.seconds_start_time + sound.relation.seconds_child_end_after_parent_start
    elif sound.relation.seconds_child_end_after_parent_end is not None:
        sound.seconds_end_time = sound.relation.parent_sound.seconds_end_time + sound.relation.seconds_child_end_after_parent_end
    return sound

class Track:
    def __init__(self, name: str = None, is_primary: bool = True, loop_until_primary_tracks_finish: bool = False):
        self._audio_segment = None
        self._sounds: List[Sound] = list()
        self.name = name
        self.is_primary = is_primary
        self.loop_until_primary_tracks_finish = loop_until_primary_tracks_finish

    def _render(self) -> AudioSegment:
        if not len(self._sounds) > 0:
            return None
        else:
            self._sounds[0].seconds_start_time = 0.0

        if len(self._sounds) == 1 and self._sounds[0].relation is None:
            return self._sounds[0]._audio_segment
        else:
            audio_in_progress = None
            for i, sound in enumerate(self._sounds):
                if sound.relation is None:
                    if sound.append_sound_after_last_one:
                        sound.seconds_start_time = audio_in_progress.duration_seconds
                else:
                    apply_relation_settings(sound=sound)
                    if audio_in_progress is None:
                        if sound.seconds_start_time > 0:
                            audio_in_progress = AudioSegment.silent((sound.duration_seconds + sound.seconds_start_time) * 1000)
                        elif sound.seconds_start_time < 0:
                            audio_in_progress = AudioSegment.silent((sound.duration_seconds - sound.seconds_start_time) * 1000)

                if sound.seconds_start_time is not None:
                    if audio_in_progress is not None:
                        audio_in_progress = audio_in_progress.overlay(sound._audio_segment, position=sound.seconds_start_time * 1000)
                    else:
                        audio_in_progress = sound._audio_segment

            return audio_in_progress

    def _export(self, filepath: str, format_type: str = "mp3"):
        audio_segment_result = self._render()
        if audio_segment_result is not None:
            audio_segment_result.export(out_f=filepath, format=format_type)

    def append_sound(self, sound: Sound, relation: Relation = None):
        if not isinstance(sound, Sound):
            raise Exception(f"You can only add Sound objects to a Track but you tried to add {sound}")

        if sound._in_use:
            raise Exception(f"The same Sound object can be used only once. If you want, you can create a new "
                            f"instance of the same Sound object, by doing   new_sound = my_sound.copy()")
        sound._in_use = True
        sound.append_sound_after_last_one = True

        self._sounds.append(sound)
        return self

    def append_sounds(self, sounds: list):
        for sound in sounds:
            self.append_sound(sound)
        return self

    def __add__(self, sound: Sound):
        self.append_sound(sound=sound)
        return self



import json
from flask import Blueprint, render_template, Response, request, jsonify

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow
from inoft_vocal_engine.web_interface.static.ui import SIDEBAR_BUTTONS
from inoft_vocal_engine.web_interface.templates.topbar import topbar_ui

organization_blueprint = Blueprint("organization", __name__, template_folder="templates")

@organization_blueprint.route("/<account_username>/<project_url>/organization")
def organization(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_organization, to_template=True
    )

def _organization(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    project_data, request_success = project_resources.team_organization_projects_dynamodb_client.get_project_data_by_account_project_id(
        account_project_id=project_resources.account_project_id
    )
    print(f"project_data = {project_data['M']}")  # todo: fix this shitty way of handling the project data retrieving
    return render_template("team-organization/gantt.html",
                           account_username=account_resources.account_username, project_url=project_resources.project_url,
                           resumed_project_data=json.dumps(project_data['M']), sidebar_buttons=SIDEBAR_BUTTONS,
                           topbar_buttons=topbar_ui.render(
                               account_username=account_resources.account_username,
                               project_url=project_resources.project_url,
                               active_button_id=topbar_ui.ID_ORGANIZATION
                           ))

@organization_blueprint.route("/<account_username>/<project_url>/organization/save", methods=["POST"])
def organization_save(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_organization_save, to_json=True
    )

def _organization_save(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    request_json_data = request.get_json()
    print(request_json_data)
    project_resources.team_organization_projects_dynamodb_client.save_project_data(
        account_project_id=project_resources.account_project_id, project_data=request_json_data)

    return jsonify({"success": True})



from typing import List

from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.databases.dynamodb.projects_text_contents_dynamodb_client import ProjectsTextContentsDynamoDbClient
from inoft_vocal_engine.import_integrations.botpress.botpress_converter import BotpressConverter


def put_new_botpress_contents_from_list(contents_list: list, project_resources: ProjectResources):
    from inoft_vocal_engine.import_integrations.botpress.utils import botpress_date_to_timestamp

    contents_list = BotpressConverter().botpress_contents_json_object_to_list_content_elements(list_all_text_elements=contents_list)
    for content in contents_list:
        from inoft_vocal_engine.databases.dynamodb.projects_text_contents_dynamodb_client import ContentDatabaseItem
        db_item = ContentDatabaseItem(elementId=content.id, stateId=0,
                                      creationTimestamp=botpress_date_to_timestamp(content.created_on),
                                      lastModificationTimestamp=botpress_date_to_timestamp(content.modified_on),
                                      crudeText=content.crude_text, characterNames=content.character_names)
        project_resources.project_text_contents_dynamodb_client.put_new_content(db_item)

def put_new_botpress_contents_from_filepath(filepath: str, project_resources: ProjectResources):
    contents_list = BotpressConverter().botpress_json_file_to_list_content_elements(filepath=filepath)
    put_new_botpress_contents_from_list(contents_list=contents_list, project_resources=project_resources)



def prettify_speech_text(text: str) -> str:
    new_text = str()
    last_char_index_to_have_been_added_new_line = 0
    for i_char, char in enumerate(text):
        if i_char > last_char_index_to_have_been_added_new_line + 115:
            if char == " ":
                new_text += (char + "/n")
                last_char_index_to_have_been_added_new_line = i_char
            else:
                new_text += char
        else:
            new_text += char
    return new_text

def to_class_name(text: str) -> str:
    import inflect

    new_text = ""
    current_digit_sequence = ""
    number_last_char_index = None
    for i, char in enumerate(text):
        # We add a - when adding numbers, so that they will be considered as separate
        # element in the class name, and that their first letter will be capitalized.
        if str(char).isdigit():
            current_digit_sequence += char
            if i+1 == len(text):
                new_text += ("-" + inflect.engine().number_to_words(current_digit_sequence))
                current_digit_sequence = ""
        else:
            if current_digit_sequence != "":
                new_text += ("-" + inflect.engine().number_to_words(current_digit_sequence))
                current_digit_sequence = ""
                number_last_char_index = i

            if number_last_char_index is not None and number_last_char_index == i:
                char = char.capitalize()
            new_text += char

    formatted_output_text = ""
    half_text_elements = new_text.replace(",", "").split("-")
    final_text_elements = list()
    for text_element in half_text_elements:
        for splitted_element in text_element.split(" "):
            final_text_elements.append(splitted_element)

    for text_element in final_text_elements:
        formatted_output_text += ((text_element[0].capitalize() if len(text_element) > 0 else "") +
                                  (text_element[1:] if len(text_element) > 1 else ""))

    return formatted_output_text



import enum

from ask_smapi_model.v1.skill.manifest import DistributionCountries
from pydantic import BaseModel
from typing import List, Optional

from pydantic.dataclasses import dataclass


class LanguageModel(BaseModel):
    language_code: str
    country_code: DistributionCountries
    name: str

class LocalInfos(BaseModel):
    language: LanguageModel
    skill_name: str
    skill_description: Optional[str] = " "
    skill_summary: Optional[str] = " "
    example_phrases: Optional[List[str]] = list()
    keywords: Optional[List[str]] = list()
    termsOfUseUrl: Optional[str] = " "
    privacyPolicyUrl: Optional[str] = " "


class AlexaLanguage(enum.Enum):
    french_france: LanguageModel = LanguageModel(language_code="fr-FR", country_code=DistributionCountries.FR, name="French France")

"""
alexa_language_codes = [
    LanguageModel(language_code="de-DE", name="Deustch"),
    LanguageModel(language_code="en-AU", name="Australian"),
    LanguageModel(language_code="en-CA", name="Canadian"),
    LanguageModel(language_code="en-GB", name="German"),
    LanguageModel(language_code="en-IN", name="Deustch"),
    LanguageModel(language_code="en-US", name="Deustch"),
    LanguageModel(language_code="es-ES", name="Espagnol Espagne"),
    LanguageModel(language_code="es-MX", name="Deustch"),
    LanguageModel(language_code="es-US", name="Deustch"),
    LanguageModel(language_code="fr-CA", name="French Canadian"),
    LanguageModel(language_code="fr-FR", name="French France"),
    LanguageModel(language_code="hi-IN", name="Hindou India"),
    LanguageModel(language_code="it-IT", name="Deustch"),
    LanguageModel(language_code="ja-JP", name="Deustch"),
    LanguageModel(language_code="pt-BR", name="Deustch")
]
"""

def try_to_convert_to_dict(object_instance: any, return_none_on_failure: bool) -> dict or any:
    if hasattr(object_instance, "to_dict"):
        return object_instance.to_dict()
    elif hasattr(object_instance, "__dict__"):
        return object_instance.__dict__
    else:
        if return_none_on_failure is False:
            return object_instance
        else:
            return None

def nested_serialize(object_instance: any):
    if isinstance(object_instance, list):
        for i in range(len(object_instance)):
            object_instance[i] = nested_serialize(object_instance=object_instance[i])
        return object_instance
    elif isinstance(object_instance, dict):
        for key, item in object_instance.items():
            object_instance[key] = nested_serialize(object_instance=item)
        return object_instance
    else:
        object_data_dict: Optional[dict] = try_to_convert_to_dict(
            object_instance=object_instance, return_none_on_failure=True
        )

        if object_data_dict is not None:
            output_data_dict = object_data_dict.copy()
            attribute_map: Optional[dict] = getattr(object_instance, "attribute_map", None)

            for key, item in object_data_dict.items():
                print(f"{key} {item}")
                if attribute_map is not None:
                    attribute_map_key = attribute_map.get(key, None)
                    if attribute_map_key is not None and key != attribute_map_key:
                        output_data_dict[attribute_map_key] = try_to_convert_to_dict(
                            object_instance=nested_serialize(object_instance.__dict__[key]),
                            return_none_on_failure=False
                        )
                        output_data_dict.pop(key)
                        continue

                current_item_to_serialize = object_instance.__dict__.get(key, None)
                if current_item_to_serialize is None:
                    # In some rare cases, like with an Enum class, trying to access the data with the __dict__ super
                    # function will change the keys of the variable, and we need to access the data via the object_data_dict.
                    current_item_to_serialize = object_data_dict.get(key, None)
                if current_item_to_serialize is not None:
                    output_data_dict[key] = try_to_convert_to_dict(
                        object_instance=nested_serialize(current_item_to_serialize),
                        return_none_on_failure=False
                    )
            return output_data_dict
        else:
            return object_instance



class ProjectDeploymentInfo(BaseModel):
    class AlexaModel(BaseModel):
        invocationName: str
        testingInstructionsText: str
        category: str
        localInfos: List[LocalInfos]
    alexa: AlexaModel

    def to_alexa_manifest(self):
        from ask_smapi_model.v1.skill.manifest import SkillManifest
        from ask_smapi_model.v1.isp import PublishingInformation
        from ask_smapi_model.v1.isp import PrivacyAndCompliance
        from ask_smapi_model.v1.skill.manifest import SkillManifestEvents
        from ask_smapi_model.v1.skill.manifest import PermissionItems
        from ask_smapi_model.v1.skill.manifest import SkillManifestApis
        from ask_smapi_model.v1.skill.alexa_hosted import HostingConfiguration
        from ask_smapi_model.v1.skill.manifest import DistributionCountries
        from ask_smapi_model.v1.skill.manifest import CustomApis
        from ask_smapi_model.v1.skill.manifest import SkillManifestEndpoint

        from ask_smapi_model.v1.skill.manifest import SkillManifestPublishingInformation
        from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPublishingInformation
        from ask_smapi_model.v1.skill.manifest import SkillManifestPrivacyAndCompliance
        alexa_manifest = SkillManifest(
            manifest_version="1.0",
            publishing_information=SkillManifestPublishingInformation(
                locales={}, distribution_countries=[],
                is_available_worldwide=False,
                testing_instructions="1) Say 'Alexa, discover my devices' 2)",
                category="SMART_HOME"
            ),
            privacy_and_compliance=SkillManifestPrivacyAndCompliance(
                locales={},
                allows_purchases=False,
                is_export_compliant=True,
                is_child_directed=False,
                uses_personal_info=False
            ),
            apis={
                "custom": {
                    "endpoint": {
                        "uri": "arn:aws:lambda:us-east-1:032174894474:function:ask-custom-custome_cert"
                    }
                }
            }
            # events=SkillManifestEvents(),
            # permissions=[PermissionItems()],
            # apis=SkillManifestApis(
            #    custom=CustomApis(
            #        endpoint=SkillManifestEndpoint(
            #            uri="arn:aws:lambda:eu-west-3:631258222318:function:testlambdacoolose"
            #        )
            #    )
            # )
        )

        for local in self.alexa.localInfos:
            local: LocalInfos
            if local.language.country_code not in alexa_manifest.publishing_information.distribution_countries:
                alexa_manifest.publishing_information.distribution_countries.append(local.language.country_code.value)

            from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPublishingInformation
            alexa_manifest.publishing_information.locales[local.language.language_code] = (
                SkillManifestLocalizedPublishingInformation(
                    name=local.skill_name,
                    description=local.skill_description,
                    summary=local.skill_summary,
                    example_phrases=local.example_phrases,
                    keywords=local.keywords
                )
            )

            from ask_smapi_model.v1.skill.manifest import SkillManifestLocalizedPrivacyAndCompliance
            alexa_manifest.privacy_and_compliance.locales[local.language.language_code] = (
                SkillManifestLocalizedPrivacyAndCompliance(
                    privacy_policy_url=local.privacyPolicyUrl,
                    terms_of_use_url=local.termsOfUseUrl
                )
            )
            # todo: check if the urls are valid http or https url, otherwise
            #  there will be an internal client error when sending the data
        # serialized_manifest = nested_serialize(alexa_manifest)
        return alexa_manifest



import socket
import time

HOST, PORT = "localhost", 9999
data = "yoooooy"

while True:
    # Create a socket (SOCK_STREAM means a TCP socket)
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        # Connect to server and send data
        sock.connect((HOST, PORT))

        sock.sendall(bytes(data + "\n", "utf-8"))
        print(f"Sent:     {data}")

        # Receive data from the server and shut down
        received = str(sock.recv(1024), "utf-8")
        print(f"Received: {received}")

        time.sleep(0.1)



import os
from typing import Optional

from inoft_vocal_framework.utils.general import load_json


class Simulator:
    PLATFORM_ALEXA = "alexa"
    PLATFORM_GOOGLE = "google"
    PLATFORM_BIXBY = "bixby"
    AVAILABLE_PLATFORMS = [PLATFORM_ALEXA, PLATFORM_GOOGLE, PLATFORM_BIXBY]

    def __init__(self, event_type: str, platform: Optional[str] = None, is_event_with_wrapper: Optional[bool] = False):
        self.platform = platform
        self.is_event_with_wrapper = is_event_with_wrapper

        self.current_dir_path = os.path.dirname(os.path.abspath(__file__))
        self.wrappers_folderpath = os.path.join(self.current_dir_path, "wrappers")
        self.request_samples_folderpath = os.path.join(self.current_dir_path, "request_samples")
        self.event_file_filepath = os.path.join(self.request_samples_folderpath, "" if platform is None else platform, f"{event_type}.json")
        if not os.path.isfile(self.event_file_filepath):
            raise Exception(f"No event file was found at {self.event_file_filepath}. Please use a valid event_type name.")

    def get_event_and_context(self) -> (str, str):
        event = None
        context = None

        if self.is_event_with_wrapper is False:
            event_body = load_json(self.event_file_filepath)

            if self.platform not in self.AVAILABLE_PLATFORMS:
                raise Exception(f"The platform named {self.platform} has not been found in the available platforms : {self.AVAILABLE_PLATFORMS}")

            if self.platform == self.PLATFORM_GOOGLE:
                event = load_json(os.path.join(self.wrappers_folderpath, "api_gateway_google_path_wrapper.json"))
                event["body"] = event_body
            elif self.platform == self.PLATFORM_ALEXA:
                event = event_body
            elif self.platform == self.PLATFORM_BIXBY:
                pass
        else:
            event = load_json(self.event_file_filepath)

        return event, context



import_bot_from_botpress_to_new_project = "{requester_account_id}:import-data:import-bot-from-botpress-to-new-project"
import_bot_from_botpress_in_existing_project = "{requester_account_id}:{account_project_id}:import-data:import-bot-from-botpress-in-existing-project"



from typing import List, Dict

from pydantic import ValidationError

from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.import_integrations.botpress.botpress_models import FlowFileBotpressModel, UiFileBotpressModel
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.import_integrations.botpress.engine_models import ContentElement, NodeDatabaseItem, \
    DatabaseTransitionConditionItem


class BotpressConverter:
    @staticmethod
    def convert_diagram_files(project_resources: ProjectResources, diagram_flow_data_dict: dict, diagram_ui_data_dict: dict):
        output_nodes_database_models: List[NodeDatabaseItem] = list()

        flow_populated_model_instance = None
        ui_populated_model_instance = None
        try:
            flow_populated_model_instance = FlowFileBotpressModel(**diagram_flow_data_dict)
            ui_populated_model_instance = UiFileBotpressModel(**diagram_ui_data_dict)
        except ValidationError as e:
            # todo: fix issues with validation errors. Instead of discaring the entire diagram when there is an
            #  error in one node, discard only the problematic node (BUT KEEP ALL THE VARIALE CHECKING !!!)
            print(e)

        if flow_populated_model_instance is not None and ui_populated_model_instance is not None:
            flow_ids_to_nodes: Dict[str, FlowFileBotpressModel.NodeModel] = dict()
            flow_names_to_nodes: Dict[str, FlowFileBotpressModel.NodeModel] = dict()
            for node in flow_populated_model_instance.nodes:
                flow_ids_to_nodes[node.id] = node
                flow_names_to_nodes[node.name] = node

            ui_ids_to_nodes: Dict[str, UiFileBotpressModel.NodePositionModel] = dict()
            for node in ui_populated_model_instance.nodes:
                ui_ids_to_nodes[node.id] = node

            for node_id, unique_node in flow_ids_to_nodes.items():
                unique_node_safedict = SafeDict(unique_node)
                try:
                    current_matching_node_position = ui_ids_to_nodes[unique_node.id]
                except Exception as e:
                    current_matching_node_position = None

                transition_conditions: List[DatabaseTransitionConditionItem] = list()
                if unique_node.next is not None:
                    for next_condition in unique_node.next:
                        # todo: check if a try except is faster than a in keys
                        current_next_condition_safedict = SafeDict(next_condition)
                        current_next_condition_target_condition = current_next_condition_safedict.get("condition").to_str(default=None)

                        current_next_condition_target_node_name = current_next_condition_safedict.get("node").to_str(default=None)
                        if current_next_condition_target_node_name in flow_names_to_nodes.keys():
                            current_next_condition_target_node_safedict = SafeDict(flow_names_to_nodes[current_next_condition_target_node_name])
                            current_next_item_outPortNodeId = current_next_condition_target_node_safedict.get("id").to_str(default=None)
                        else:
                            current_next_item_outPortNodeId = None

                        transition_conditions.append(DatabaseTransitionConditionItem(condition=current_next_condition_target_condition,
                                                                         targetNodeId=current_next_item_outPortNodeId))

                output_nodes_database_models.append(NodeDatabaseItem(
                    accountId=project_resources.account_username,  # todo: change incorrect use of account_username to account_id
                    accountProjectId=project_resources.account_project_id,
                    accountProjectInstanceId="12",
                    nodeId=unique_node.id,
                    nodeName=unique_node.name,
                    nodeType=unique_node.type,
                    actions=unique_node.onEnter,
                    transitions=transition_conditions,
                    temp_on_receive=unique_node.onReceive,
                    positionX=None if current_matching_node_position is None else current_matching_node_position.position.x,
                    positionY=None if current_matching_node_position is None else current_matching_node_position.position.y
                ))

        return output_nodes_database_models


    """"@staticmethod
    def botpress_node_to_engine_database_node(botpress_node: ):"""

    @staticmethod
    def botpress_contents_json_object_to_list_content_elements(list_all_text_elements: list) -> List[ContentElement]:
        from inoft_vocal_engine.inoft_vocal_markup.deserializer import Deserializer
        inoft_vocal_markup_deserializer = Deserializer(characters_names=["Léo", "Willie", "Menu"])

        content_elements: List[ContentElement] = list()
        for text_element in list_all_text_elements:
            current_content_element = ContentElement()
            current_content_element.character_names = list()

            text_element_safedict = SafeDict(text_element)
            text_content = text_element_safedict.get("formData").get("text$fr").to_str(default=None)
            if text_content is not None:
                current_content_element.crude_text = text_content
                current_content_element.dialogues_lines = inoft_vocal_markup_deserializer.deserialize(text=text_content)
                # todo: give the ability to select the language to deserialize

                for dialogue_line in current_content_element.dialogues_lines:
                    if dialogue_line.character_name is not None:
                        if dialogue_line.character_name not in current_content_element.character_names:
                            current_content_element.character_names.append(dialogue_line.character_name)
            else:
                current_content_element.crude_text = " "

            current_content_element.id = text_element_safedict.get("id").to_str(default=None)
            current_content_element.created_by = text_element_safedict.get("createdBy").to_str(default=None)
            current_content_element.created_on = text_element_safedict.get("createdOn").to_str(default=None)
            current_content_element.modified_on = text_element_safedict.get("modifiedOn").to_str(default=None)
            content_elements.append(current_content_element)

        return content_elements

    @staticmethod
    def botpress_json_file_to_list_content_elements(filepath: str) -> List[ContentElement]:
        from inoft_vocal_framework.utils.general import load_json
        return BotpressConverter.botpress_contents_json_object_to_list_content_elements(list_all_text_elements=load_json(filepath))



from flask import Flask, Response, render_template, request

from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.route_blueprints.auth import auth_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.audio_editor import audio_editor_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.code_editor import code_editor_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.contents import contents_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.control_panel import control_panel_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.deployment.deployment import deployment_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.diagrams import diagrams_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.intents import intents_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.dynamic_url_for import dynamic_url_for
from inoft_vocal_engine.web_interface.route_blueprints.permissions import permissions_blueprint
from inoft_vocal_engine.web_interface.route_blueprints.team_organization import organization_blueprint

def strip_spaces(string: str):
    return string.strip()

app = Flask(__name__)
app.config["shouldUseCloudDist"] = False
app.config["bucketUrl"] = "https://inoft-vocal-engine-web-test.s3.eu-west-3.amazonaws.com"
app.config["bucketDirPath"] = "prod"
app.jinja_env.globals.update(url_for=dynamic_url_for)
# app.config['SERVER_NAME'] = 'example.com'
app.jinja_env.filters["strip_spaces"] = strip_spaces

app.register_blueprint(control_panel_blueprint)
app.register_blueprint(contents_blueprint)
app.register_blueprint(audio_editor_blueprint)
app.register_blueprint(organization_blueprint)
app.register_blueprint(diagrams_blueprint)
app.register_blueprint(intents_blueprint)
app.register_blueprint(auth_blueprint)
app.register_blueprint(code_editor_blueprint)
app.register_blueprint(permissions_blueprint)
app.register_blueprint(deployment_blueprint)


@app.route("/")
def index():
    from inoft_vocal_engine.web_interface.route_blueprints.contents import contents
    return contents("anvers1944")

@app.after_request
def add_header(r):
    """
    Add headers to both force latest IE rendering engine or Chrome Frame,
    and also to cache the rendered page for 10 minutes.
    """
    r.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
    r.headers["Pragma"] = "no-cache"
    r.headers["Expires"] = "0"
    r.headers['Cache-Control'] = 'public, max-age=0'
    return r

@app.route("/text")
def text():
    return render_template("text-editor/test.html")

@app.route("/equalizer")
def equalizer():
    return render_template("equalizer/index.html")

@app.route("/project_dir", methods=["POST"])
def change_project_dir():
    data = SafeDict(request.get_json())
    filepath = data.get("filepath").to_str(default=None)
    # list_content = content_list(filepath=filepath)
    # return jsonify({"html": list_content})

@app.errorhandler(404)
def error_404(error):
    return Response("Erreur 404")

def lambda_handler(event, context):
    import awsgi
    print(f"event :")
    print(event)
    print("context :")
    print(context)
    return awsgi.response(app=app, event=event, context=context)

if __name__ == '__main__':
    # import serverless_wsgi
    # rest_data = {'resource': '/', 'path': '/', 'httpMethod': 'GET', 'headers': None, 'multiValueHeaders': None, 'queryStringParameters': None, 'multiValueQueryStringParameters': None, 'pathParameters': None, 'stageVariables': None, 'requestContext': {'resourceId': 's1dd4la1h9', 'resourcePath': '/', 'httpMethod': 'GET', 'extendedRequestId': 'Qy0WcGnbiGYFZmQ=', 'requestTime': '05/Aug/2020:11:39:59 +0000', 'path': '/', 'accountId': '631258222318', 'protocol': 'HTTP/1.1', 'stage': 'test-invoke-stage', 'domainPrefix': 'testPrefix', 'requestTimeEpoch': 1596627599596, 'requestId': 'a79afdd7-17e7-47e9-820b-cc2990626e6e', 'identity': {'cognitoIdentityPoolId': None, 'cognitoIdentityId': None, 'apiKey': 'test-invoke-api-key', 'principalOrgId': None, 'cognitoAuthenticationType': None, 'userArn': 'arn:aws:iam::631258222318:root', 'apiKeyId': 'test-invoke-api-key-id', 'userAgent': 'aws-internal/3 aws-sdk-java/1.11.820 Linux/4.9.217-0.1.ac.205.84.332.metal1.x86_64 OpenJDK_64-Bit_Server_VM/25.252-b09 java/1.8.0_252 vendor/Oracle_Corporation', 'accountId': '631258222318', 'caller': '631258222318', 'sourceIp': 'test-invoke-source-ip', 'accessKey': 'ASIAZF6PMB3XEKJLVGOT', 'cognitoAuthenticationProvider': None, 'user': '631258222318'}, 'domainName': 'testPrefix.testDomainName', 'apiId': 'bfcq7086f9'}, 'body': None, 'isBase64Encoded': False}
    # print(serverless_wsgi.handle_request(app=app, event=rest_data, context=None))
    app.run()



import socketserver

class MyTCPHandler(socketserver.BaseRequestHandler):
    """
    The request handler class for our server.

    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """

    def handle(self):
        # self.request is the TCP socket connected to the client
        self.data = self.request.recv(1024).strip()
        print(f"{self.client_address[0]} wrote")
        print(self.data)
        # just send back the same data, but upper-cased
        self.request.sendall(self.data.upper())

if __name__ == "__main__":
    HOST, PORT = "localhost", 9999

    # Create the server, binding to localhost on port 9999
    server = socketserver.TCPServer((HOST, PORT), MyTCPHandler)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    server.serve_forever()



import json
from flask import Blueprint, render_template, request, Response, jsonify
from pydantic import ValidationError
from typing import List

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.databases.dynamodb.utils import remove_null_elements_from_dict
from inoft_vocal_engine.safe_dict import SafeDict
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow

diagrams_blueprint = Blueprint("diagrams", __name__, template_folder="templates")

@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams")
def diagrams(account_username: str, project_url: str):
    from inoft_vocal_engine.permissions.diagrams import project_diagrams_nodes_read_all
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[project_diagrams_nodes_read_all],
        success_handler=_diagrams, to_template=True,
    )

def _diagrams(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    response = project_resources.project_diagrams_data_dynamodb_client.query_nodes_by_project_id(
        project_id=project_resources.account_project_id
    )

    from inoft_vocal_engine.import_integrations.botpress.engine_models import NodeDatabaseItem
    nodes_data_dicts: List[dict] = list()
    for item in response.items:
        if isinstance(item, dict):
            current_node_database_item = NodeDatabaseItem(**item)
            nodes_data_dicts.append(current_node_database_item.to_processed_node_element().dict())

    return render_template("diagrams/index.html", nodes_data=json.dumps(nodes_data_dicts),
                           **project_resources.required_template_kwargs())

# todo: centralize the imports and import the bot all at once
@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/import/botpress/bot", methods=["POST"])
def import_botpress_bot(account_username: str, project_url: str):
    zip_file = request.data
    binary_file_path = 'file.zip'  # Name for new zip file you want to regenerate
    with open(binary_file_path, 'wb') as f:
        f.write(zip_file)
    input()

# todo: rename this function to import/botpress
@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/upload", methods=["POST"])
def upload_diagrams_data(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_upload_diagrams_data, to_json=True,
    )

def _upload_diagrams_data(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    request_json_data = SafeDict(request.get_json())
    diagram_data = request_json_data.get("diagramData").to_safedict()
    # todo: save the rest of the fields of the diagram data (version, catchAll, startNode, description)
    #  in a separate database, like the accountDiagramsDatabaseClient
    node_elements = diagram_data.get("nodes").to_list(default=None)
    if node_elements is not None:
        from inoft_vocal_engine.import_integrations.botpress.diagrams_data import put_new_botpress_diagrams_data_from_list
        put_new_botpress_diagrams_data_from_list(nodes_list=node_elements,
            projects_diagrams_data_dynamodb_client=project_resources.project_diagrams_data_dynamodb_client)

    generate_code(project_resources=project_resources)
    # todo: create a button on the ui that can generated the code, and create a build system

    return Response(status=200)

perm = "{account_username}:{project_id}:diagrams:nodes:update:all"

# todo: rename this function to import/botpress
@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/nodes/update", methods=["POST"])
def update_nodes(account_username: str, project_url: str):
    from inoft_vocal_engine.permissions.diagrams import project_diagrams_nodes_update_all
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[project_diagrams_nodes_update_all],
        success_handler=_update_nodes, to_json=True,
    )

def _update_nodes(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    nodes_data = SafeDict(request.get_json()).get("nodesData").to_list()
    print(nodes_data)

    from inoft_vocal_engine.import_integrations.botpress.engine_models import NodeElementDataReceivedFromUser
    for node in nodes_data:
        try:
            node_element = NodeElementDataReceivedFromUser(**node)
            project_resources.project_diagrams_data_dynamodb_client.update_node(node_element_data=node_element)
        except ValidationError as e:
            print(e)

    return jsonify(success=True)

@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/nodes/add", methods=["POST"])
def add_nodes(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_add_nodes, to_json=True,
    )

def _add_nodes(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    nodes_data = SafeDict(request.get_json()).get("nodesData").to_list()
    print(nodes_data)

    from inoft_vocal_engine.import_integrations.botpress.engine_models import NodeElementDataReceivedFromUser, NodeDatabaseItem
    for node in nodes_data:
        try:
            node_element = NodeElementDataReceivedFromUser(**node)
            node_database_item = NodeDatabaseItem(
                accountId=project_resources.project_owner_account_id,
                accountProjectId=project_resources.account_project_id,
                accountProjectInstanceId="null",
                **remove_null_elements_from_dict(node_element.dict())
            )
            project_resources.project_diagrams_data_dynamodb_client.put_new_node(node_database_item)
        except ValidationError as e:
            print(e)

    return Response(status=200)

@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/links/add", methods=["POST"])
def add_links(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_add_links, to_json=True,
    )

def _add_links(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    links_data = SafeDict(request.get_json()).to_list()

    from inoft_vocal_engine.import_integrations.botpress.engine_models import DatabaseTransitionConditionItem
    for link in links_data:
        try:
            transition_condition = DatabaseTransitionConditionItem(**link)
            project_resources.project_diagrams_data_dynamodb_client.add_update_transition_to_node(
                transition_condition=transition_condition
            )
        except ValidationError as e:
            print(e)

    return jsonify(success=True)



@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/actions/create/one", methods=["POST"])
def create_one_action(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_create_one_action, to_json=True,
    )

def _create_one_action(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.models.actions import SayTextAction
    try:
        # todo: add support for multiple types of actions

        say_text_action = SayTextAction.make_new(text_content_id=None, **request.get_json())
        # Create a new SayTextAction to validate the arguments, and initialize the action id.

        from inoft_vocal_engine.inoft_vocal_markup.deserializer import Deserializer
        inoft_vocal_markup_deserializer = Deserializer(characters_names=["Léo", "Willie", "Menu"])
        # Create a markup deserializer, that will be passed to the ContentDatabaseItem, so that we can get the
        # dialogues lines of the crudeText, which will allow us to get the names of the characters in a dialogue.
        # todo: make creation of deserializer dynamic based on character names stored in a database

        from inoft_vocal_engine.databases.dynamodb.projects_text_contents_dynamodb_client import ContentDatabaseItem
        text_content_item = ContentDatabaseItem.make_new(
            crude_text=say_text_action.loadedTextToSay,
            markup_deserializer=inoft_vocal_markup_deserializer,
            ids_instances_using_content=["say_text_action.instanceId"]
            # todo: add instanceId in the BaseAction class
        )
        project_resources.project_text_contents_dynamodb_client.put_new_content(text_content_item)
        # Once created, put the new content item in the database.

        say_text_action.textContentItemId = text_content_item.elementId
        # Once its done, we cna assign the content item id to the say_text_action. We needed to create the
        # say_text_action element before the text_content_item, in order to access the action instanceId,
        # after the data as been correctly validated by passing it in the SayTextAction class.
        response = project_resources.project_diagrams_data_dynamodb_client.add_update_action_to_node(action_instance=say_text_action)
        # Then, we can simply add the action to the node.

        say_text_action.isNew = False
        # We set the isNew value on False, so that when we return the new object that will be load in the
        # client browser, the action will not be considered as new, and will now be updated instead of created.

        if response is not None:
            return jsonify(success=True, action=say_text_action.dict())
    except ValidationError as e:
        print(e)
    return jsonify(success=False)

@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/actions/update/one", methods=["POST"])
def update_one_action(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_update_one_action, to_json=True,
    )

def _update_one_action(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.models.actions import SayTextAction
    try:
        say_text_action = SayTextAction(**request.get_json())
        print(say_text_action)
        response = project_resources.project_diagrams_data_dynamodb_client.add_update_action_to_node(action_instance=say_text_action)

        # todo: update text element if the action is a SayTextAction

        return jsonify(success=True)
    except ValidationError as e:
        print(e)
        return jsonify(success=False)

@diagrams_blueprint.route("/<account_username>/<project_url>/diagrams/create-new", methods=["POST"])
def create_new_diagram(account_username: str, project_url: str):
    # todo: add required_permissions
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_create_new_diagram, to_json=True,
    )

def _create_new_diagram(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    from inoft_vocal_engine.databases.dynamodb.accounts_data_dynamodb_client import InstanceInfoDatabaseItem
    request_data = SafeDict(request.get_json())

    instance_name = request_data.get("instanceName").to_str(default=None)
    if instance_name is None:
        return jsonify({"success": False})
    else:
        instance_description = request_data.get("instanceDescription").to_str(default=None)
        operation_success = engine_resources.accounts_data_dynamodb_client.add_instance_infos_to_account(
            account_id=project_resources.project_owner_account_id,
            instance_infos_item=InstanceInfoDatabaseItem(
                projectId=project_resources.account_project_id,
                instanceName=instance_name, instanceDescription=instance_description
            )
        )
        return jsonify({"success": operation_success})





from pydantic import BaseModel, StrictStr
from typing import List


class AudioProject(BaseModel):
    class _Collections(BaseModel):
        class _Tracks(BaseModel):
            class _Model(BaseModel):
                class _Attributes(BaseModel):
                    class _File(BaseModel):
                        name: str
                        webkitRelativePath: str
                        lastModified: int
                        size: int
                        type: str

                    class _Buffer(BaseModel):
                        duration: float
                        length: int
                        sampleRate: int
                        numberOfChannels: int

                    file: _File
                    buffer: _Buffer
                    color: str
                    length: float
                    name: str
                    solo: bool
                    pan: int
                    muted: bool
                    gain: float
                attributes: _Attributes
            models: List[_Model]
        tracks: _Tracks
    projectId: StrictStr
    collections: _Collections

    def add_track(self, color: str = "e", gain: float = 1.0, pan: int = 0, solo: bool = False, muted: bool = False):
        dummy = {}
        self.collections.tracks.models.append(
            self._Collections._Tracks._Model(
                attributes=self._Collections._Tracks._Model._Attributes(
                    file=self._Collections._Tracks._Model._Attributes._File(
                        name="e", webkitRelativePath="e", lastModified=0, size=0, type="e"
                    ),
                    buffer=self._Collections._Tracks._Model._Attributes._Buffer(
                        duration=0.0, length=1, sampleRate=1, numberOfChannels=2
                    ),
                    color=color, length=100.0, name="Default", solo=solo, pan=pan, muted=muted, gain=gain
                )
            )
        )


"""
{
  'projectId': 'builtin_text-lbs0Re',
  'Collections': {
    'Tracks': {
      'models': [{
        'attributes': {
          'file': {
            'name': 'jean sablon - alexa.mp3',
            'webkitRelativePath': '',
            'lastModified': 1591264478848.0,
            'size': 1007568.0,
            'type': 'audio/mpeg'
          },
          'color': '#00a0b0',
          'length': 1920.0,
          'name': 'Track 1',
          'solo': False,
          'buffer': {
            'duration': 167.71,
            'length': 7396188.0,
            'sampleRate': 44100.0,
            'numberOfChannels': 2.0
          },
          'pan': 0,
          'muted': False,
          'gain': 1.0
        }
      }]
    }
  }
} & text_project_data = {
 
"""


from typing import Optional, Any, List
from boto3.dynamodb.conditions import Key, Attr
from boto3.exceptions import ResourceNotExistsError
from pydantic import BaseModel, ValidationError

from inoft_vocal_engine.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, GlobalSecondaryIndex, PrimaryIndex, \
    Response
from inoft_vocal_engine.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_engine.inoft_vocal_markup.deserializer import DialogueLine, Deserializer

import StructNoSQL


class ContentDatabaseItem(BaseModel):
    elementId: str
    creationTimestamp: int
    lastModificationTimestamp: int
    stateId: int
    crudeText: str
    characterNames: Optional[list] = list()
    idsInstancesUsingContent: Optional[list] = list()

    @classmethod
    def make_new(cls, crude_text: str, markup_deserializer: Deserializer,
                 ids_instances_using_content: Optional[List[str]] = None):
        from uuid import uuid4
        element_id = str(uuid4())

        character_names: List[str] = list()
        dialogues_lines = markup_deserializer.deserialize(text=crude_text)
        for dialogue_line in dialogues_lines:
            if dialogue_line.character_name is not None:
                if dialogue_line.character_name not in character_names:
                    character_names.append(dialogue_line.character_name)

        import time
        current_time = int(time.time())
        return cls(
            elementId=element_id, creationTimestamp=current_time, lastModificationTimestamp=current_time,
            stateId=0, crudeText=crude_text, characterNames=character_names,
            idsInstancesUsingContent=ids_instances_using_content
        )

class ProjectsTextContentsDynamoDbClient(DynamoDbCoreAdapter):
    """
    State Ids :
    -2 = Deleted but stored
    -1 = Archived, and will not be classically displayed
    0 = Available in the standard interface
    """

    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="elementId", hash_key_variable_python_type=str)

        globals_secondary_indexes = [
            # GlobalSecondaryIndex(index_name="characterNames", key_type=GlobalSecondaryIndex.KEY_TYPE_IS_HASH_KEY,
            #                     variable_python_type=list, projection_type=GlobalSecondaryIndex.PROJECTION_TYPE_USE_ALL),
            # I cannot have a sort key be a list, so i guess to get all the content that include a character, with a scan query :/
            GlobalSecondaryIndex(hash_key_name="instanceId", hash_key_variable_python_type=int,
                                 sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="instanceId", hash_key_variable_python_type=int,
                                 sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="stateId", hash_key_variable_python_type=int,
                                 sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="stateId", hash_key_variable_python_type=int,
                                 sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL")
        ]
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index,
                         create_table=True, global_secondary_indexes=globals_secondary_indexes)

    def put_new_content(self, db_item: ContentDatabaseItem):
        try:
            table = self.dynamodb.Table(self.table_name)
            table.put_item(Item=db_item.dict())
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def update_content_element(self, element_id: str, dialogue_line_index: int, element_text: str) -> Response:
        # todo: add support for dialogue line, because currently we are replacing the lines by a single text
        try:
            table = self.dynamodb.Table(self.table_name)
            dialogue_line = DialogueLine(character_name="Default", line_content=element_text, additional_character_metadata=None)
            # Todo: add support for character name and additional character metadata

            from time import time
            response = table.update_item(
                Key={"elementId": element_id},
                UpdateExpression=f"SET dialogueLines[{dialogue_line_index}]=:dialogueLine, lastModificationTimestamp=:timestamp",
                ExpressionAttributeValues={
                    ':dialogueLine': dialogue_line.dict(),
                    ":timestamp": round(time())
                    # We need the timestamp to be an int, not a float, so we round it.
                },
                ReturnValues="UPDATED_NEW"
            )
            print(response)
            return response

        except ResourceNotExistsError:
            raise Exception( f"DynamoDb table {self.table_name} doesn't exist. Failed to update attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to update attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def _get_latest(self, index_name: str, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        query = {
            "TableName": self.table_name,
            "IndexName": index_name,
            "KeyConditionExpression": Key("stateId").eq(0),
            "ScanIndexForward": False,
            "Limit": num_latest_items
        }
        if exclusive_start_key is not None:
            query["ExclusiveStartKey"] = exclusive_start_key

        table = self.dynamodb.Table(self.table_name)
        response = table.query(**query)

        if "Items" in response:
            return Response(Utils.dynamodb_to_python(dynamodb_object=response))  # , text_decoding_type="utf-8"))
            # todo: check if there are no issues when i do not specify the text decoding type
        else:
            return Response({})

    def get_latest_created(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-creationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    def get_latest_updated(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-lastModificationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    """
    def put_item(self, project_data: dict) -> bool:
        try:
            print(f"Saving project data {project_data}")
            item = Utils.python_to_dynamodb(project_data)
            self.dynamodb.put_item(TableName=self.table_name, Item=item)
            return True
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save project data to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")
    """

    def get_by_element_id(self, element_id: str) -> Optional[ContentDatabaseItem]:
        response = self._get_item_by_primary_key(
            key_name="elementId", key_value=element_id,
            fields_to_get=["characterNames", "crudeText"]
        )
        if response is not None and response.success is True:
            try:
                content_item = ContentDatabaseItem(**response.item)
                return content_item
            except ValidationError as e:
                print(e)
        return None

    def get_by_character_name(self, character_name: str) -> Response:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.scan(FilterExpression=Attr('characterNames').contains(character_name))
            # todo: change that to a query instead of a scan
            if "Items" in response:
                return Response(Utils.dynamodb_to_python(dynamodb_object=response))
            else:
                return Response({})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")

    def get_by_text_search(self, text_to_search: str, section_instance_id: int) -> Response:
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.query(KeyConditionExpression=Key("instanceId").eq(section_instance_id),
                                   IndexName="instanceId-creationTimestamp",
                                   FilterExpression=Attr('crudeText').contains(text_to_search))
            if "Items" in response:
                return Response(Utils.dynamodb_to_python(dynamodb_object=response))
            else:
                return Response({})
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")


class ProjectsTextContentsTableDataModel(StructNoSQL.TableDataModel):
    element_id = StructNoSQL.BaseField(name="elementId", field_type=str)
    creation_timestamp = StructNoSQL.BaseField(name="creationTimestamp", field_type=int)  # StructNoSQL.unixTimestamp
    last_modification_timestamp = StructNoSQL.BaseField(name="lastModificationTimestamp", field_type=int)  # StructNoSQL.unixTimestamp
    crude_text = StructNoSQL.BaseField(name="crudeText", field_type=str)
    character_names = StructNoSQL.BaseField(name="characterNames", field_type=list)  # List[str]
    section_instance_id = StructNoSQL.BaseField(name="sectionInstanceId", field_type=int)
    state_id = StructNoSQL.BaseField(name="stateId", field_type=int)

class ProjectsTextContentsDynamoDBNewTableClient(StructNoSQL.BaseTable):
    def __init__(self, table_name: str, region_name: str):
        primary_index = StructNoSQL.PrimaryIndex(hash_key_name="elementId", hash_key_variable_python_type=str)

        globals_secondary_indexes = [
            # StructNoSQL.GlobalSecondaryIndex(index_name="characterNames", key_type=GlobalSecondaryIndex.KEY_TYPE_IS_HASH_KEY,
            #                     variable_python_type=list, projection_type=GlobalSecondaryIndex.PROJECTION_TYPE_USE_ALL),
            # I cannot have a sort key be a list, so i guess to get all the content that include a character, with a scan query :/
            StructNoSQL.GlobalSecondaryIndex(
                hash_key_name="instanceId", hash_key_variable_python_type=int,
                sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                projection_type="ALL"),
            StructNoSQL.GlobalSecondaryIndex(
                hash_key_name="instanceId", hash_key_variable_python_type=int,
                sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                projection_type="ALL"),
            StructNoSQL.GlobalSecondaryIndex(
                hash_key_name="stateId", hash_key_variable_python_type=int,
                sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                projection_type="ALL"),
            StructNoSQL.GlobalSecondaryIndex(
                hash_key_name="stateId", hash_key_variable_python_type=int,
                sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                projection_type="ALL")
        ]
        super().__init__(
            table_name=table_name, region_name=region_name,
            data_model=ProjectsTextContentsTableDataModel(), auto_create_table=True,
            primary_index=primary_index, global_secondary_indexes=globals_secondary_indexes
        )


if __name__ == "__main__":
    table_client = ProjectsTextContentsDynamoDBNewTableClient(
        table_name="inoft-vocal-engine_accounts-projects-instances_text-contents-data", region_name="eu-west-2"
    )
    table_client.model: ProjectsTextContentsTableDataModel
    # table_client.query()

    """
    projects_text_contents_dynamodb_static_client = ProjectsTextContentsDynamoDbClient(
        table_name="inoft-vocal-engine_accounts-projects-instances_text-contents-data", region_name="eu-west-2"
    )
    # data = projects_text_contents_dynamodb_static_client.get_latest_updated(num_latest_items=3).items
    # data = projects_text_contents_dynamodb_static_client.get_by_character_name(character_name="Léo").items
    data = projects_text_contents_dynamodb_static_client.get_by_text_search(
        text_to_search="bruits", section_instance_id=12).items
    print(data)
    """



from typing import Optional

from flask import Blueprint, render_template, request, jsonify
from pydantic import ValidationError

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.smapi_clients.smapi_base_client import SmapiBaseClient
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow
from inoft_vocal_engine.web_interface.route_blueprints.deployment.models import ApplicationManifestUpdateRequestData, \
    ApplicationManifestGetRequestData, switch_key_type_to_data_models

deployment_blueprint = Blueprint("deployment", __name__, template_folder="templates")

@deployment_blueprint.route("/<account_username>/<project_url>/deployment")
def deployment(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_deployment, to_template=True,
    )

def _deployment(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    return render_template(
        "deployment/deployment.html",
        **project_resources.required_template_kwargs()
    )


@deployment_blueprint.route("/<account_username>/<project_url>/publishing/get", methods=["POST"])
def get_publishing_infos(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_get_publishing_infos, to_json=True,
    )

def _get_publishing_infos(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    try:
        request_data = ApplicationManifestGetRequestData(**request.get_json())
        validators_class_types = switch_key_type_to_data_models.get(request_data.typeKeyName, None)
        if validators_class_types is not None:
            validator_class_instance = validators_class_types.base(**request_data.data)

            response_item_dict: Optional[dict] = project_resources.vocal_apps_model_schemas_dynamodb_client.get_item_in_path_target(
                account_project_id=project_resources.account_project_id,
                target_path_elements=validator_class_instance.database_path,
            )
            if response_item_dict is not None:
                sender_validator_class_instance = validators_class_types.sender(**response_item_dict)
                return jsonify({"success": True, "data": sender_validator_class_instance.json()})
    except ValidationError as e:
        print(e)
    return jsonify({"success": False})

@deployment_blueprint.route("/<account_username>/<project_url>/publishing/update", methods=["POST"])
def update_publishing_infos(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_update_publishing_infos, to_json=True,
    )

def _update_publishing_infos(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    try:
        request_data = ApplicationManifestUpdateRequestData(**request.get_json())
        validators_class_types = switch_key_type_to_data_models.get(request_data.typeKeyName, None)
        if validators_class_types is not None:
            validator_class_instance = validators_class_types.update(**request_data.data)

            response = project_resources.vocal_apps_model_schemas_dynamodb_client.update_value_with_path_target(
                account_project_id=project_resources.account_project_id,
                object_path_elements=validator_class_instance.database_path,
                value=validator_class_instance.export()
            )
            return jsonify({"success": True})
    except ValidationError as e:
        print(e)
    return jsonify({"success": False})


"""

        refresh_token = "Atzr|IwEBIJPMVNBwbBGY2hum-1Q0ldBPx4wXyFiPr38X9F1Y4FgPYjHvTLK5UlsG5z4kXZlMZHIYGDQLtFFw7saqZ4ZEaRalmjLfdR62uBXe_PupGVrrsX3721TZrSG68Om7VvBCFqJS3BHM_AG5b9VFJkDIkmO5wz45vo9S-6JR1I2KLB12N_vGSKZ5SfIma1PSHjfeAeQgWBQfIa19KE0xBw8t9U0fyGaM-JvwxoLulf1k265hWUDDj3Uuve5CbUlU41NHtFYZXbVrj76vEGIw7-HZBoKU0ZA1-_oscarmnZ5b5ENFFIuKt-bGbQJw5ycgT0Gm0Ja3sxGI7EpakcGBIOECOl0Am4QZBCBRtRENvHVHKtjhT2PeZEZh9Re_DB4nRnnpJOXXXiwajfNdcvtSPHP9nOXtg64iaALnZBeZzzzhcVEXsqQh9TEwO1Y6nOGFMtDm-zOFulGjDwvz_ySLmA5qGbjz9o4p_-A5DEXG2zIgezPT0Z_SyqsiVq7GJ3ixAydyG_4R7nSGEo63_tt34Ffuq4-eSV605Rt7nEHUivm5M9QFOUargug98MlZJala-z_x1u5ETJAZpVhxkXfl4b-XMbqA1YkW2wfO9IkJf_P2UzX2rY1TPz9nGn_DUiVSeGucePo"
        # todo: make refresh token dynamic (store the refresh token of an user in the database after he logged with
        #  amazon, and if the token is not found, or was not working, prompt the user to login with amazon)
        smapi_client = SmapiBaseClient(refresh_token=refresh_token)
        smapi_client.create_skill(vendor_id=smapi_client.get_vendors_of_account()[0].id,
                                  skill_manifest=publishing_infos.to_alexa_manifest())
        # todo: instead of getting the first vendor of the account, get the main vendor of the account
"""


from typing import Optional, Any, List
from boto3.dynamodb.conditions import Key
from boto3.exceptions import ResourceNotExistsError
from pydantic import BaseModel

from inoft_vocal_framework.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, GlobalSecondaryIndex, PrimaryIndex, \
    Response
from inoft_vocal_framework.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_framework.inoft_vocal_markup.deserializer import DialogueLine


class ContentItem(BaseModel):
    elementId: str
    creationTimestamp: int
    lastModificationTimestamp: int
    sectionInstanceId: int
    stateId: int
    dialogueLines: Optional[List[DialogueLine]] = None

class ProjectsTextContentsDynamoDbClient(DynamoDbCoreAdapter):
    """
    State Ids :
    -2 = Deleted but stored
    -1 = Archived, and will not be classically displayed
    0 = Available in the standard interface
    """

    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="elementId", hash_key_variable_python_type=str)

        globals_secondary_indexes = [
            # GlobalSecondaryIndex(index_name="characterNames", key_type=GlobalSecondaryIndex.KEY_TYPE_IS_HASH_KEY,
            #                     variable_python_type=list, projection_type=GlobalSecondaryIndex.PROJECTION_TYPE_USE_ALL),
            # I cannot have a sort key be a list, so i guess to get all the content that include a character, with a scan query :/
            GlobalSecondaryIndex(hash_key_name="sectionInstanceId", hash_key_variable_python_type=int,
                                 sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="sectionInstanceId", hash_key_variable_python_type=int,
                                 sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="stateId", hash_key_variable_python_type=int,
                                 sort_key_name="creationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL"),
            GlobalSecondaryIndex(hash_key_name="stateId", hash_key_variable_python_type=int,
                                 sort_key_name="lastModificationTimestamp", sort_key_variable_python_type=int,
                                 projection_type="ALL")
        ]
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index,
                         create_table=True, global_secondary_indexes=globals_secondary_indexes)

    def put_new_content(self, db_item: ContentItem):
        try:
            table = self.dynamodb.Table(self.table_name)
            table.put_item(Item=db_item.dict())
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def update_content_element(self, element_id: str, dialogue_line_index: int, element_text: str) -> Response:
        # todo: add support for dialogue line, because currently we are replacing the lines by a single text
        try:
            table = self.dynamodb.Table(self.table_name)
            dialogue_line = DialogueLine(character_name="Default", line_content=element_text, additional_character_metadata=None)
            # Todo: add support for character name and additional character metadata

            from time import time
            response = table.update_item(
                Key={"elementId": element_id},
                UpdateExpression=f"SET dialogueLines[{dialogue_line_index}]=:dialogueLine, lastModificationTimestamp=:timestamp",
                ExpressionAttributeValues={
                    ':dialogueLine': dialogue_line.dict(),
                    ":timestamp": round(time())
                    # We need the timestamp to be an int, not a float, so we round it.
                },
                ReturnValues="UPDATED_NEW"
            )
            print(response)
            return response

        except ResourceNotExistsError:
            raise Exception( f"DynamoDb table {self.table_name} doesn't exist. Failed to update attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to update attributes to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def _get_latest(self, index_name: str, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        query = {
            "TableName": self.table_name,
            "IndexName": index_name,
            "KeyConditionExpression": Key("stateId").eq(0),
            "ScanIndexForward": False,
            "Limit": num_latest_items
        }
        if exclusive_start_key is not None:
            query["ExclusiveStartKey"] = exclusive_start_key

        table = self.dynamodb.Table(self.table_name)
        response = table.query(**query)

        # We use a scan instead of a query, since we just want the latest items without any conditions.
        if "Items" in response:
            return Response(Utils.dynamodb_to_python(dynamodb_object=response))  # , text_decoding_type="utf-8"))
            # todo: check if there are no issues when i do not specify the text decoding type
        else:
            return Response({})

    def get_latest_created(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-creationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    def get_latest_updated(self, num_latest_items: int, exclusive_start_key: Optional[dict] = None) -> Response:
        return self._get_latest(index_name="stateId-lastModificationTimestamp",
                                num_latest_items=num_latest_items, exclusive_start_key=exclusive_start_key)

    """
    def put_item(self, project_data: dict) -> bool:
        try:
            print(f"Saving project data {project_data}")
            item = Utils.python_to_dynamodb(project_data)
            self.dynamodb.put_item(TableName=self.table_name, Item=item)
            return True
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save project data to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")
    """

    def get_by_id(self, project_id: str) -> (dict, bool):
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.get_item(TableName=self.table_name, Key={"elementId": project_id}, ConsistentRead=True)
            if "Item" in response:
                return Utils.dynamodb_to_python(dynamodb_object=response["Item"]), True
            else:
                return dict(), False
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")


if __name__ == "__main__":
    projects_text_contents_dynamodb_static_client = ProjectsTextContentsDynamoDbClient(
        table_name="test-inoft-vocal-engine-project-text-contents", region_name="eu-west-2"
    )
    data = projects_text_contents_dynamodb_static_client.get_latest_updated(num_latest_items=3).items



import os
from pathlib import Path
from typing import Optional, List, Dict
from uuid import uuid4
from zipfile import ZipFile

from pydantic.dataclasses import dataclass

from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.import_integrations.botpress.botpress_converter import BotpressConverter
from inoft_vocal_engine.utils.general import load_json
from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path, get_engine_temp_path


@dataclass
class DiagramData:
    flow_data_dict: dict
    ui_data_dict: dict


class BotpressImportCore(BotpressConverter):
    def __init__(self, existing_extracted_archive_dirpath: Optional[str] = None,
                 source_zip_filepath: Optional[str] = None, source_zip_file_bytes: Optional[bytes] = None):
        self._existing_extracted_archive_dirpath = existing_extracted_archive_dirpath
        self._source_zip_filepath = source_zip_filepath
        self._source_zip_file_bytes = source_zip_file_bytes
        self._source_zip_filename = Path(self._source_zip_filepath).stem

        if self._existing_extracted_archive_dirpath is not None:
            self.extracted_archive_dirpath = self._existing_extracted_archive_dirpath
        else:
            self.extracted_archive_dirpath = os.path.join(get_engine_temp_path(), f"{self._source_zip_filename}_{uuid4()}")
            self._load_zip()
        self.extracted_archive_root_dirpath = os.path.join(self.extracted_archive_dirpath, self._source_zip_filename)

        self.flows_dirpath = os.path.join(self.extracted_archive_root_dirpath, "flows")

    def _load_zip(self):
        if self._source_zip_filepath is not None:
            bot_root_archive = ZipFile(self._source_zip_filepath)
            bot_root_archive.extractall(path=self.extracted_archive_dirpath)
        elif self._source_zip_file_bytes is not None:
            pass
        else:
            raise Exception(f"source_zip_filepath and source_zip_file_bytes cannot be both to None."
                            f"  --source_zip_filepath:{self._source_zip_filepath}"
                            f"  --source_zip_file_bytes:{self._source_zip_file_bytes}")

    def get_all_diagrams_data_from_zip(self) -> List[DiagramData]:
        diagrams_list: List[DiagramData] = list()

        if os.path.exists(self.flows_dirpath):
            for root, dirs, files in os.walk(self.flows_dirpath):
                for name in files:
                    current_filepath = os.path.join(root, name)
                    current_filepath_and_its_suffixes = current_filepath.split(".")
                    current_filepath_without_suffixes = current_filepath_and_its_suffixes[0]
                    current_filepath_suffixes = current_filepath_and_its_suffixes[1:]

                    if "json" in current_filepath_suffixes:
                        expected_flow_filepath = None
                        expected_ui_filepath = None

                        if "flow" in current_filepath_suffixes:
                            expected_flow_filepath = current_filepath
                            expected_ui_filepath = f"{current_filepath_without_suffixes}.ui.json"
                        elif "ui" in current_filepath_suffixes:
                            expected_ui_filepath = current_filepath
                            expected_flow_filepath = f"{current_filepath_without_suffixes}.flow.json"

                        if os.path.isfile(expected_flow_filepath) and os.path.isfile(expected_ui_filepath):
                            flow_data_dict = load_json(filepath=expected_flow_filepath)
                            ui_data_dict = load_json(filepath=expected_ui_filepath)
                            try:
                                diagrams_list.append(DiagramData(flow_data_dict=flow_data_dict, ui_data_dict=ui_data_dict))
                            except Exception as e:
                                print(f"{e} -- Error while trying to add new diagram with"
                                      f"  --flow_data_dict:{flow_data_dict}"
                                      f"  --ui_data_dict:{ui_data_dict}")

        return diagrams_list


if __name__ == "__main__":
    core = BotpressImportCore(
        existing_extracted_archive_dirpath="F:/Inoft/anvers_1944_project/inoft_vocal_engine/tmp/anvers_1944_787027a5-2aff-448c-979a-1cbfcffd9b3f",
        source_zip_filepath="C:/Users/LABOURDETTE/Documents/Software/BotPress/botpress-v12_8_1-win-x64/data/bots/anvers_1944.zip"
    )
    project_resources = ProjectResources(engine_resources=EngineResources(),
                                         account_username="robinsonlabourdette",
                                         project_url="anvers1944")

    diagrams_data = core.get_all_diagrams_data_from_zip()
    for diagram in diagrams_data:
        output = core.convert_diagram_files(project_resources=project_resources,
                                            diagram_flow_data_dict=diagram.flow_data_dict,
                                            diagram_ui_data_dict=diagram.ui_data_dict)
        print(output)
        for node in output:
            project_resources.project_diagrams_data_dynamodb_client.put_new_node(node_item=node)





from boto3.exceptions import ResourceNotExistsError
from inoft_vocal_framework.databases.dynamodb.dynamodb_core import DynamoDbCoreAdapter, PrimaryIndex
from inoft_vocal_framework.databases.dynamodb.dynamodb_utils import Utils


class TeamOrganizationProjectsDynamoDbClient(DynamoDbCoreAdapter):
    def __init__(self, table_name: str, region_name: str):
        primary_index = PrimaryIndex(hash_key_name="projectId", hash_key_variable_python_type=str)
        super().__init__(table_name=table_name, region_name=region_name, primary_index=primary_index, create_table=True)

    def save_project_data(self, project_id: str, project_data: dict) -> bool:
        try:
            item = Utils.python_to_dynamodb(project_data)
            item["projectId"] = project_id
            print(f"Saving team organization project data {item}")
            table = self.dynamodb.Table(self.table_name)
            table.put_item(Item=item)
            return True
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} doesn't exist. Failed to save attributes to DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to save project data to DynamoDb table. Exception of type {type(e).__name__} occurred: {str(e)}")

    def get_project_data_by_project_id(self, project_id: str) -> (dict, bool):
        # If the value from get_field is of dict or list type, the SafeDict will be populated, otherwise it will be empty without errors.
        try:
            table = self.dynamodb.Table(self.table_name)
            response = table.get_item(Key={"projectId": project_id}, ConsistentRead=True)
            if "Item" in response:
                # Utils.decimal_deserializer(
                return Utils.dynamodb_to_python(response["Item"]), True
            else:
                return dict(), False
        except ResourceNotExistsError:
            raise Exception(f"DynamoDb table {self.table_name} do not exist or in the process"
                            "of being created. Failed to get attributes from DynamoDb table.")
        except Exception as e:
            raise Exception(f"Failed to retrieve attributes from DynamoDb table."
                            f"Exception of type {type(e).__name__} occurred: {str(e)}")



import time
from typing import Optional, Tuple

import botocore
from botocore.exceptions import ClientError

# todo: fix issue where when deploying we do not check if the api with the id found in the settings file do exit
from inoft_vocal_engine.cloud_providers.aws.aws_core import AwsCore
from inoft_vocal_engine.cloud_providers.aws.generated_code_s3_client import GeneratedCodeS3Client
from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.databases.dynamodb.accounts_data_dynamodb_client import AWSBuildInfosBackendModel
from inoft_vocal_engine.utils.paths import get_inoft_vocal_engine_root_path
from inoft_vocal_framework.utils.general import generate_uuid4


class DeployHandler(AwsCore):
    def __init__(self):
        print("Initializing AWS clients... (this can take a few seconds)")
        start_time = time.time()
        super().__init__(clients_to_load=[AwsCore.CLIENT_S3, AwsCore.CLIENT_LAMBDA, AwsCore.CLIENT_API_GATEWAY,
                                          AwsCore.CLIENT_IAM, AwsCore.CLIENT_STS, AwsCore.CLIENT_BOTO_SESSION])
        print(f"Took {round(time.time() - start_time, 2)}s to initiate AWS clients")
        # todo: make the initialization even more asynchronous, by calling the handle function while the AWS ressources are initializing.
        #  Right now, the initialization of the resources are async, but we must wait for their completion in order to call the handle function.

    def deploy_deprecated(self, lambda_files_root_folderpath: str, stage_name: str, bucket_name: str, bucket_region_name: str,
               lambda_name: str, lambda_handler: str, runtime="python3.7", upload_zip: bool = True,
               existing_zip_filepath: Optional[str] = None, lambda_description: str = "Inoft Vocal Engine Deployment",
               lambda_timeout_seconds: int = 30, lambda_memory_size: int = 512, publish: bool = True):
        pass

    def _try_to_update_lambda_code(self, lambda_arn: str, bucket_name: str, object_key_name: str, handler_function_path: str) -> bool:
        try:
            self.update_lambda_function_code(lambda_arn=lambda_arn, object_key_name=object_key_name, bucket_name=bucket_name)
            self.update_lambda_function_configuration(function_name=lambda_arn, handler_function_path=handler_function_path)
            print(f"Updated the code of the lambda with arn {lambda_arn}")
            return True
        except botocore.exceptions.ClientError as e:
            print(e)
            return False

    def deploy(self, engine_resources: EngineResources, project_resources: ProjectResources,
               lambda_handler: str, runtime="python3.7", lambda_description: str = "Inoft Vocal Engine Deployment",
               lambda_timeout_seconds: int = 30, lambda_memory_size: int = 512, publish: bool = True) -> AWSBuildInfosBackendModel:

        has_modified_build_data = False

        # Retrieve or initialize the build data.
        build_data: Optional[AWSBuildInfosBackendModel] = engine_resources.accounts_data_dynamodb_client.get_infos_of_one_project_build(
            project_resources=project_resources, build_id="1d52c25e-bd69-4a66-95d5-23b0c747d974"
        )
        if build_data is None:
            build_data = AWSBuildInfosBackendModel(buildId=generate_uuid4())
            has_modified_build_data = True

        # Variables that will be used multiple times in the function
        expected_build_s3_key = f"{project_resources.project_owner_account_id}/{project_resources.account_project_id}/builds/build1.zip"
        expected_lambda_function_arn = f"{project_resources.account_project_id}-build-{build_data.buildName}"

        # Create and configure the Lambda function
        if build_data.lambdaArn is not None:
            if self.lambda_function_exist(function_name_or_arn=build_data.lambdaArn) is not True:
                build_data.lambdaArn = None
                has_modified_build_data = True

        # We do not use an elif statement here, to handle an non-existent lambda arn in the build data.
        if build_data.lambdaArn is None:
            if self.lambda_function_exist(function_name_or_arn=expected_lambda_function_arn) is True:
                build_data.lambdaArn = expected_lambda_function_arn
                has_modified_build_data = True

        if build_data.lambdaArn is not None:
            success = self._try_to_update_lambda_code(
                lambda_arn=build_data.lambdaArn, bucket_name=GeneratedCodeS3Client.PROJECTS_DATA_V1_BUCKET_NAME,
                object_key_name=expected_build_s3_key, handler_function_path=lambda_handler
            )
        else:
            # After creating the lambda function, we do not need to update
            # its code, because we defined its code while creating it.
            arn_created_lambda = self.create_lambda_function(
                bucket=GeneratedCodeS3Client.PROJECTS_DATA_V1_BUCKET_NAME, s3_key=expected_build_s3_key,
                function_name=f"{project_resources.account_project_id}-build-{build_data.buildName}",
                handler=lambda_handler, description=lambda_description,
                timeout=lambda_timeout_seconds, memory_size=lambda_memory_size,
                runtime=runtime, publish=publish
            )
            if arn_created_lambda is not None:
                build_data.lambdaArn = arn_created_lambda
                has_modified_build_data = True

        # Create and configure the API Gateway
        if build_data.apiGatewayId is not None:
            if self.rest_api_gateway_exist(build_data.apiGatewayId) is not True:
                build_data.apiGatewayId = None
                has_modified_build_data = True

        if build_data.apiGatewayId is not None:
            print(f"Using an existing API Gateway for deployment."
                  f"\n  --apiGatewayId:{build_data.apiGatewayId}")
        else:
            if build_data.lambdaArn is not None:
                api_id = self.create_api_gateway(lambda_arn=build_data.lambdaArn, api_gateway_name="the-api-noice", route_names=['{proxy+}'])
                if api_id is not None:
                    build_data.apiGatewayId = api_id
                    has_modified_build_data = True
                    print(f"Created a new API Gateway during deployment."
                          f"\n  --apiId:{api_id}")

        if has_modified_build_data is True:
            database_update_build_data_success = engine_resources.accounts_data_dynamodb_client.set_update_infos_of_one_project_build(
                project_resources=project_resources, build=build_data
            )

        return build_data


if __name__ == "__main__":
    engine_resources = EngineResources()
    account_resources = AccountResources(engine_resources=engine_resources)
    account_resources.account_username = "robinsonlabourdette"
    project_resources = ProjectResources(engine_resources=engine_resources, account_resources=account_resources,
                                         project_url="anvers1944", project_owner_account_username="robinsonlabourdette",
                                         project_owner_account_id="b1fe5939-032b-462d-92e0-a942cd445096")

    DeployHandler().deploy(engine_resources=engine_resources, project_resources=project_resources, lambda_handler="diagrams.lambda_handler")



from typing import Optional


class DynamoDBTable:
    def __init__(self, db_table_name: str, db_region_name: str):
        self.db_table_name = db_table_name
        self.db_region_name = db_region_name

class DatabaseSessionsUsersData(DynamoDBTable):
    def __init__(self, db_table_name: str = None, db_region_name: str = None, disable_database: Optional[bool] = False):
        super().__init__(db_table_name=db_table_name, db_region_name=db_region_name)
        self.disable_database = disable_database
        if self.disable_database is not True:
            if self.db_table_name is None:
                pass
                # prompt_database_warning_message(variable_name="db_table_name", instance_type=DatabaseSessionsUsersData)
            if self.db_region_name is None:
                pass
                # prompt_database_warning_message(variable_name="db_region_name", instance_type=DatabaseSessionsUsersData)


class DatabaseMessagesContent(DynamoDBTable):
    def __init__(self, db_table_name: str = None, db_region_name: str = None, disable_database: Optional[bool] = False):
        super().__init__(db_table_name=db_table_name, db_region_name=db_region_name)
        self.disable_database = disable_database

class DatabaseUsersNotificationsSubscriptions(DynamoDBTable):
    def __init__(self, db_table_name: str = None, db_region_name: str = None, disable_database: Optional[bool] = False):
        super().__init__(db_table_name=db_table_name, db_region_name=db_region_name)
        self.disable_database = disable_database



from flask import Blueprint, render_template, jsonify, request
from typing import List

from inoft_vocal_engine.cloud_ressources.account_resources import AccountResources
from inoft_vocal_engine.cloud_ressources.engine_resources import EngineResources
from inoft_vocal_engine.cloud_ressources.project_resources import ProjectResources
from inoft_vocal_engine.web_interface.auth.backend import auth_workflow
from inoft_vocal_engine.web_interface.static.ui import SIDEBAR_BUTTONS
from inoft_vocal_engine.web_interface.templates.topbar import topbar_ui

code_editor_blueprint = Blueprint("code-editor", __name__, template_folder="templates")

@code_editor_blueprint.route("/<account_username>/<project_url>/code-editor")
def code_editor(account_username: str, project_url: str):
    return auth_workflow(
        account_username=account_username, project_url=project_url,
        required_permissions=[],
        success_handler=_code_editor, to_template=True,
    )

def _code_editor(engine_resources: EngineResources, account_resources: AccountResources, project_resources: ProjectResources):
    generated_files = _generate_application_code(project_resources=project_resources)

    # from inoft_vocal_engine.cloud_providers.aws.generated_code_s3_client import GeneratedCodeS3Client
    # app_code = GeneratedCodeS3Client().get_app_code()

    return render_template(
        "code-editor/index.html", app_code=generated_files[0],
        sidebar_buttons=SIDEBAR_BUTTONS,
        topbar_buttons=topbar_ui.render(
            account_username=project_resources.project_owner_account_username,
            project_url=project_resources.project_url,
            active_button_id=topbar_ui.ID_ORGANIZATION
        ),
        **project_resources.required_template_kwargs()
    )

def _generate_application_code(project_resources: ProjectResources) -> List[str]:
    from inoft_vocal_engine.code_generation.generator import GeneratorCore
    code_generated_files = GeneratorCore(project_resources=project_resources).generate()

    from inoft_vocal_engine.cloud_providers.aws.generated_code_s3_client import GeneratedCodeS3Client
    s3_client = GeneratedCodeS3Client()

    # todo: make it possible to have custom filename, instead of just diagrams_{i}
    for i, file in enumerate(code_generated_files):
        s3_client.upload_app_code(project_resources=project_resources, app_code=file, key_filename=f"diagrams.py")

    return code_generated_files

"""
@audio_editor_blueprint.route("/audio-editor/<account_project_id>/save", methods=["POST"])
def audio_editor_save(account_project_id: str):
    print(f"Saving data for project {account_project_id}")

    request_json_data = request.get_json()
    if request_json_data is not None:
        request_success = project_resources.audio_editor_projects_dynamodb_client.save_project_data(account_project_id=account_project_id,
                                                                                                    project_data=request_json_data)
    else:
        request_success = False

    return jsonify({"success": request_success})
"""




from decimal import Decimal

import boto3
from boto3.session import Session
from typing import List, Optional, Type

from pydantic import BaseModel, StrictStr, validate_arguments

from inoft_vocal_framework.databases.dynamodb.dynamodb_utils import Utils
from inoft_vocal_framework.safe_dict import SafeDict
from inoft_vocal_framework.utils.static_logger import logger

HASH_KEY_TYPE = "HASH"
SORT_KEY_TYPE = "RANGE"


class Response:
    def __init__(self, response_dict: dict):
        response_safedict = SafeDict(response_dict)
        self.items = response_safedict.get("Items").to_list(default=None)
        self.count = response_safedict.get("Count").to_int(default=None)
        self.scanned_count = response_safedict.get("ScannedCount").to_int(default=None)
        self.last_evaluated_key = response_safedict.get("LastEvaluatedKey").to_dict(default=None)
        self.has_reached_end = False if self.last_evaluated_key is not None else True

class Index(BaseModel):
    hash_key_name: str
    hash_key_variable_python_type: Type
    sort_key_name: Optional[str]
    sort_key_variable_python_type: Optional[Type]
    index_custom_name: Optional[str]


class PrimaryIndex(Index):
    pass

class GlobalSecondaryIndex(Index):
    # todo: add pydantic to this class
    PROJECTION_TYPE_USE_ALL = "ALL"
    PROJECTION_TYPE_KEYS_ONLY = "KEYS_ONLY"
    PROJECTION_TYPE_INCLUDE = "INCLUDE"
    ALL_PROJECTIONS_TYPES = [PROJECTION_TYPE_USE_ALL, PROJECTION_TYPE_KEYS_ONLY, PROJECTION_TYPE_INCLUDE]

    projection_type: str

    def __init__(self, hash_key_name: str, hash_key_variable_python_type: Type,
                 sort_key_name: Optional[str], sort_key_variable_python_type: Optional[Type], projection_type: str,
                 index_custom_name: Optional[str] = None):

        super().__init__(hash_key_name=hash_key_name, hash_key_variable_python_type=hash_key_variable_python_type,
                         sort_key_name=sort_key_name, sort_key_variable_python_type=sort_key_variable_python_type,
                         projection_type=projection_type)

        if projection_type not in self.ALL_PROJECTIONS_TYPES:
            raise Exception(f"{projection_type} has not been found in the available projection_types : {self.ALL_PROJECTIONS_TYPES}")
        self.projection_type = projection_type

    def to_dict(self):
        if self.index_custom_name is not None:
            index_name = self.index_custom_name
        else:
            if self.sort_key_name is None:
                index_name = self.hash_key_name
            else:
                index_name = f"{self.hash_key_name}-{self.sort_key_name}"

        output_dict = {
            "IndexName": index_name,
            "KeySchema": [
                {
                    "AttributeName": self.hash_key_name,
                    "KeyType": HASH_KEY_TYPE
                },
            ],
            "Projection": {
                "ProjectionType": self.projection_type
            }
        }
        if self.sort_key_name is not None and self.sort_key_variable_python_type is not None:
            output_dict["KeySchema"].append({
                "AttributeName": self.sort_key_name,
                "KeyType": SORT_KEY_TYPE
            })
        return output_dict


class CreateTableQueryKwargs:
    @validate_arguments
    def __init__(self, table_name: str, billing_mode: str):
        self._names_already_defined_attributes: List[str] = list()
        self.data = {
            "TableName": table_name,
            "KeySchema": list(),
            "AttributeDefinitions": list(),
            "BillingMode": billing_mode,
        }

    def _add_key(self, key_name: str, key_python_variable_type: Type, key_type: str):
        self.data["KeySchema"].append({
            "AttributeName": key_name,
            "KeyType": key_type
        })
        self.data["AttributeDefinitions"].append({
            "AttributeName": key_name,
            "AttributeType": Utils.python_type_to_dynamodb_type(python_type=key_python_variable_type)
        })

    @validate_arguments
    def add_hash_key(self, key_name: str, key_python_variable_type: Type):
        self._add_key(key_name=key_name, key_python_variable_type=key_python_variable_type, key_type=HASH_KEY_TYPE)

    @validate_arguments
    def add_sort_key(self, key_name: str, key_python_variable_type: Type):
        self._add_key(key_name=key_name, key_python_variable_type=key_python_variable_type, key_type=SORT_KEY_TYPE)

    def _add_global_secondary_index(self, key_name: str, key_python_variable_type: Type):
        if key_name not in self._names_already_defined_attributes:
            self.data["AttributeDefinitions"].append({
                "AttributeName": key_name,
                "AttributeType": Utils.python_type_to_dynamodb_type(python_type=key_python_variable_type)
            })
            self._names_already_defined_attributes.append(key_name)

    @validate_arguments
    def add_global_secondary_index(self, global_secondary_index: GlobalSecondaryIndex):
        if "GlobalSecondaryIndexes" not in self.data:
            self.data["GlobalSecondaryIndexes"] = list()

        self._add_global_secondary_index(key_name=global_secondary_index.hash_key_name,
                                         key_python_variable_type=global_secondary_index.hash_key_variable_python_type)
        if global_secondary_index.sort_key_name is not None and global_secondary_index.sort_key_variable_python_type is not None:
            self._add_global_secondary_index(key_name=global_secondary_index.sort_key_name,
                                             key_python_variable_type=global_secondary_index.sort_key_variable_python_type)

        self.data["GlobalSecondaryIndexes"].append(global_secondary_index.to_dict())

    @validate_arguments
    def add_all_global_secondary_indexes(self, global_secondary_indexes: List[GlobalSecondaryIndex]):
        for global_secondary_index in global_secondary_indexes:
            self.add_global_secondary_index(global_secondary_index=global_secondary_index)


if __name__ == "__main__":
    c = CreateTableQueryKwargs(table_name="test", billing_mode="e")
    c.add_hash_key(key_name="t", key_python_variable_type=int)


class DynamoDbCoreAdapter:
    _EXISTING_DATABASE_CLIENTS = dict()
    PAY_PER_REQUEST = "PAY_PER_REQUEST"
    PROVISIONED = "PROVISIONED"

    def __init__(self, table_name: str, region_name: str, primary_index: PrimaryIndex,
                 create_table: bool = True, billing_mode: str = PAY_PER_REQUEST,
                 global_secondary_indexes: List[GlobalSecondaryIndex] = None):
        self.table_name = table_name
        self.primary_index = primary_index
        self.create_table = create_table
        self.billing_mode = billing_mode
        self.global_secondary_indexes = global_secondary_indexes

        # We store the database clients in a static variable, so that if we init the class with
        # the same region_name, we do not need to wait for a new initialization of the client.
        if region_name in self._EXISTING_DATABASE_CLIENTS.keys():
            self.dynamodb = self._EXISTING_DATABASE_CLIENTS[region_name]
            print(f"Re-using the already created dynamodb client for region {region_name}")
        elif "default" in self._EXISTING_DATABASE_CLIENTS.keys():
            self.dynamodb = self._EXISTING_DATABASE_CLIENTS["default"]
            print(f"Re-using the already created dynamodb client for the default region")
        else:
            print(f"Initializing the {self}. For local development, make sure that you are connected to internet."
                  f"\nOtherwise the framework will get stuck at initializing the {self}")

            dynamodb_regions = Session().get_available_regions("dynamodb")
            if region_name in dynamodb_regions:
                self.dynamodb = boto3.resource("dynamodb", region_name=region_name)
                self._EXISTING_DATABASE_CLIENTS[region_name] = self.dynamodb
            else:
                self.dynamodb = boto3.resource("dynamodb")
                self._EXISTING_DATABASE_CLIENTS["default"] = self.dynamodb
                logger.debug(f"Warning ! The specified dynamodb region_name {region_name} is not a valid region_name."
                             f"The dynamodb client has been initialized without specifying the region.")

        self._create_table_if_not_exists()
        print(f"Initialization of {self} completed successfully !")

    def _create_table_if_not_exists(self) -> None:
        """
        Creates table in Dynamodb resource if it doesn't exist and create_table is set as True.
        :raises: PersistenceException: When `create_table` fails on dynamodb resource.
        """
        if self.create_table:
            create_table_query_kwargs = CreateTableQueryKwargs(table_name=self.table_name, billing_mode=self.billing_mode)

            create_table_query_kwargs.add_hash_key(key_name=self.primary_index.hash_key_name,
                                                   key_python_variable_type=self.primary_index.hash_key_variable_python_type)

            if self.primary_index.sort_key_name is not None and self.primary_index.sort_key_variable_python_type is not None:
                create_table_query_kwargs.add_hash_key(key_name=self.primary_index.sort_key_name,
                                                       key_python_variable_type=self.primary_index.sort_key_variable_python_type)

            if self.global_secondary_indexes is not None:
                create_table_query_kwargs.add_all_global_secondary_indexes(global_secondary_indexes=self.global_secondary_indexes)
            try:
                print(create_table_query_kwargs.data)
                self.dynamodb.create_table(**create_table_query_kwargs.data)
            except Exception as e:
                if e.__class__.__name__ != "ResourceInUseException":
                    raise Exception(f"Create table if not exists request failed: Exception of type {type(e).__name__} occurred {str(e)}")





from ask_smapi_model.v1.skill import SkillSummary
from ask_smapi_model.v1.skill.manifest import SkillManifest
from ask_smapi_sdk import StandardSmapiClientBuilder
from typing import List
from ask_smapi_model.v1.vendor_management import Vendors, Vendor
from inoft_vocal_engine.STATIC_DATA import StaticData



class SmapiBaseClient:
    def __init__(self, refresh_token: str):
        self._client_id = StaticData().ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_ID
        self._client_secret = StaticData().ALEXA_INOFT_VOCAL_ENGINE_APPLICATION_CLIENT_SECRET

        self.smapi_client_builder = StandardSmapiClientBuilder(
            client_id=self._client_id,
            client_secret=self._client_secret,
            refresh_token=refresh_token
        )
        self.smapi_client = self.smapi_client_builder.client()

    def get_vendors_of_account(self) -> List[Vendor]:
        output_vendors: List[Vendor] = list()
        vendors_response: Vendors = self.smapi_client.get_vendor_list_v1()
        if vendors_response is not None:
            for vendor in vendors_response.vendors:
                output_vendors.append(vendor)
        return output_vendors

    def get_list_skills_of_account(self) -> List[SkillSummary]:
        from ask_smapi_model.v1.skill import ListSkillResponse
        output_skills: List[SkillSummary] = list()

        vendors = self.get_vendors_of_account()
        if len(vendors) > 0:
            response: ListSkillResponse = self.smapi_client.list_skills_for_vendor_v1(vendor_id=vendors[0].id)
            output_skills.extend(response.skills)

        return output_skills

    def create_skill_legacy(self, vendor_id: str):
        from inoft_vocal_engine.models.projects.project_deployment_infos import ProjectDeploymentInfo
        from inoft_vocal_engine.models.projects.project_deployment_infos import LocalInfos
        from inoft_vocal_engine.models.projects.project_deployment_infos import AlexaLanguage

        alexa_manifest = ProjectDeploymentInfo(alexa=ProjectDeploymentInfo.AlexaModel(
            invocationName="Le test",
            testingInstructionsText="Do nothing",
            category="SMART_HOME",
            localInfos=[
                LocalInfos(
                    language=AlexaLanguage.french_france.value,
                    skill_name="Le super test de ouf",
                    skill_description="Rien",
                    skill_summary="Rien",
                    example_phrases=["Hey", "Hoy"],
                    keywords=["One", "Two"],
                    termsOfUseUrl="http://www.termsofuse.sampleskill.com",
                    privacyPolicyUrl="http://www.myprivacypolicy.sampleskill.com"
                )
            ]
        )).to_alexa_manifest()

        alexa_manifest = {
          'manifestVersion': '1.0',
          'publishingInformation': {
            'locales': {
              'de-DE': {
                'name': "Th",
                'summary': '',
                'description': '',
                'examplePhrases': [],
                'keywords': []
              },
              'fr-FR': {
                'name': 'aeea',
                'summary': '',
                'description': 'aeaeea',
                'examplePhrases': [],
                'keywords': []
              }
            },
            'isAvailableWorldwide': False,
            'testingInstructions': 'aeeaeeaea',
            'category': 'ALARMS_AND_CLOCKS',
            'distributionCountries': []
          },
          'privacyAndCompliance': {
            'locales': {
              'de-DE': {
                'privacyPolicyUrl': '',
                'termsOfUseUrl': ''
              },
              'fr-FR': {
                'privacyPolicyUrl': 'http://www.termsofuse.sampleskill.com',
                'termsOfUseUrl': 'http://www.myprivacypolicy.sampleskill.com'
              }
            },
            'allowsPurchases': False,
            'usesPersonalInfo': False,
            'isChildDirected': False,
            'isExportCompliant': True
          },
          'apis': {
            'custom': {
              'endpoint': {
                'uri': 'arn:aws:lambda:us-east-1:032174894474:function:ask-custom-custome_cert'
              }
            }
          }
        }

        from ask_smapi_model.v1.skill import CreateSkillRequest
        self.smapi_client.create_skill_for_vendor_v1(create_skill_request=CreateSkillRequest(
            vendor_id=vendor_id, manifest=alexa_manifest))
        """"{
                  'apis': {
                    'custom': {
                      'endpoint': {
                        'uri': 'arn:aws:lambda:us-east-1:032174894474:function:ask-custom-custome_cert'
                      }
                    }
                  },
                  'manifest_version': '1.0',
                  'privacy_and_compliance': {
                    'locales': {
                    },
                  },
                  'publishing_information': {
                    'category': 'ALARMS_AND_CLOCKS',
                    'description': "eeee",
                    'distribution_countries': [
                      'FR'
                    ],
                    'locales': {

                    },
                    'name': "Le test de l'inoft vocal engine",
                    'testing_instructions': 'Le truc de ouf'
                  }
                }
        ))"""

    def create_skill(self, vendor_id: str, skill_manifest: SkillManifest):
        from ask_smapi_model.v1.skill import CreateSkillRequest
        response = self.smapi_client.create_skill_for_vendor_v1(create_skill_request=CreateSkillRequest(
            vendor_id=vendor_id, manifest=skill_manifest,
        ))
        return response


if __name__ == "__main__":
    refresh_token = "Atzr|IwEBIJPMVNBwbBGY2hum-1Q0ldBPx4wXyFiPr38X9F1Y4FgPYjHvTLK5UlsG5z4kXZlMZHIYGDQLtFFw7saqZ4ZEaRalmjLfdR62uBXe_PupGVrrsX3721TZrSG68Om7VvBCFqJS3BHM_AG5b9VFJkDIkmO5wz45vo9S-6JR1I2KLB12N_vGSKZ5SfIma1PSHjfeAeQgWBQfIa19KE0xBw8t9U0fyGaM-JvwxoLulf1k265hWUDDj3Uuve5CbUlU41NHtFYZXbVrj76vEGIw7-HZBoKU0ZA1-_oscarmnZ5b5ENFFIuKt-bGbQJw5ycgT0Gm0Ja3sxGI7EpakcGBIOECOl0Am4QZBCBRtRENvHVHKtjhT2PeZEZh9Re_DB4nRnnpJOXXXiwajfNdcvtSPHP9nOXtg64iaALnZBeZzzzhcVEXsqQh9TEwO1Y6nOGFMtDm-zOFulGjDwvz_ySLmA5qGbjz9o4p_-A5DEXG2zIgezPT0Z_SyqsiVq7GJ3ixAydyG_4R7nSGEo63_tt34Ffuq4-eSV605Rt7nEHUivm5M9QFOUargug98MlZJala-z_x1u5ETJAZpVhxkXfl4b-XMbqA1YkW2wfO9IkJf_P2UzX2rY1TPz9nGn_DUiVSeGucePo"
    smapi_client = SmapiBaseClient(refresh_token=refresh_token)
    skills_summaries = smapi_client.get_list_skills_of_account()

    smapi_client.create_skill_legacy(vendor_id=smapi_client.get_vendors_of_account()[0].id)
